from __future__ import annotations

import asyncio
import base64
import json
import logging
import os
import re
import secrets
import shutil
import signal
import struct
import sys
from collections.abc import Iterable, Mapping, MutableMapping, Sequence
from dataclasses import dataclass
from decimal import Decimal
from functools import partial
from http import HTTPStatus
from io import StringIO
from pathlib import Path
from subprocess import CalledProcessError
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Final,
    FrozenSet,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
    override,
)
from uuid import UUID

import aiofiles
import aiohttp
import aiotools
import pkg_resources
import zmq
import zmq.asyncio
from aiodocker.docker import Docker, DockerContainer
from aiodocker.exceptions import DockerError
from aiodocker.types import PortInfo
from aiomonitor.task import preserve_termination_log
from aiotools import TaskGroup
from async_timeout import timeout

from ai.backend.common.cgroup import get_cgroup_mount_point
from ai.backend.common.data.image.types import ScannedImage
from ai.backend.common.docker import (
    MAX_KERNELSPEC,
    MIN_KERNELSPEC,
    ImageRef,
    KernelFeatures,
    LabelName,
)
from ai.backend.common.dto.agent.response import PurgeImageResp, PurgeImagesResp
from ai.backend.common.dto.manager.rpc_request import PurgeImagesReq
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.events.kernel import KernelLifecycleEventReason
from ai.backend.common.exception import ImageNotAvailable, InvalidImageName, InvalidImageTag
from ai.backend.common.json import (
    dump_json,
    load_json,
)
from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
from ai.backend.common.types import (
    AgentId,
    AutoPullBehavior,
    BinarySize,
    ClusterInfo,
    ClusterSSHPortMapping,
    ContainerId,
    ContainerStatus,
    DeviceId,
    DeviceName,
    ImageCanonical,
    ImageConfig,
    ImageRegistry,
    KernelCreationConfig,
    KernelId,
    MountPermission,
    MountTypes,
    ResourceGroupType,
    ResourceSlot,
    Sentinel,
    ServicePort,
    SessionId,
    SlotName,
    current_resource_slots,
)
from ai.backend.common.utils import (
    AsyncFileWriter,
    current_loop,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.logging.formatter import pretty

from ..agent import (
    ACTIVE_STATUS_SET,
    AbstractAgent,
    AbstractKernelCreationContext,
    ScanImagesResult,
)
from ..config.unified import AgentUnifiedConfig, ContainerSandboxType, ScratchType
from ..exception import ContainerCreationError, UnsupportedResource
from ..fs import create_scratch_filesystem, destroy_scratch_filesystem
from ..kernel import AbstractKernel
from ..plugin.network import ContainerNetworkCapability, ContainerNetworkInfo, NetworkPluginContext
from ..proxy import DomainSocketProxy, proxy_connection
from ..resources import (
    AbstractComputePlugin,
    ComputerContext,
    KernelResourceSpec,
    Mount,
    known_slot_types,
)
from ..scratch import create_loop_filesystem, destroy_loop_filesystem
from ..server import get_extra_volumes
from ..types import (
    AgentEventData,
    Container,
    KernelOwnershipData,
    LifecycleEvent,
    MountInfo,
    Port,
)
from ..utils import (
    closing_async,
    container_pid_to_host_pid,
    get_kernel_id_from_container,
    get_safe_ulimit,
    host_pid_to_container_pid,
    update_nested_dict,
)
from .kernel import DockerKernel
from .metadata.server import MetadataServer
from .resources import load_resources, scan_available_resources
from .utils import PersistentServiceContainer

if TYPE_CHECKING:
    from ai.backend.common.auth import PublicKey
    from ai.backend.common.etcd import AsyncEtcd

log = BraceStyleAdapter(logging.getLogger(__spec__.name))
eof_sentinel = Sentinel.TOKEN

LDD_GLIBC_REGEX = re.compile(r"^ldd \([^\)]+\) ([\d\.]+)$")
LDD_MUSL_REGEX = re.compile(r"^musl libc .+$")

known_glibc_distros: Final[dict[float, str]] = {
    2.17: "centos7.6",
    2.27: "ubuntu18.04",
    2.28: "centos8.0",
    2.31: "ubuntu20.04",
    2.34: "centos9.0",
    2.35: "ubuntu22.04",
    2.39: "ubuntu24.04",
}


def container_from_docker_container(src: DockerContainer) -> Container:
    ports = []
    for private_port, host_ports in src["NetworkSettings"]["Ports"].items():
        private_port = int(private_port.split("/")[0])
        if host_ports is None:
            host_ip = "127.0.0.1"
            host_port = 0
        else:
            host_ip = host_ports[0]["HostIp"]
            host_port = int(host_ports[0]["HostPort"])
        ports.append(Port(host_ip, private_port, host_port))
    return Container(
        id=ContainerId(src.id),
        status=src["State"]["Status"],
        image=src["Config"]["Image"],
        labels=src["Config"]["Labels"],
        ports=ports,
        backend_obj=src,
    )


async def _clean_scratch(
    loop: asyncio.AbstractEventLoop,
    scratch_type: str,
    scratch_root: Path,
    kernel_id: KernelId,
) -> None:
    scratch_dir = scratch_root / str(kernel_id)
    tmp_dir = scratch_root / f"{kernel_id}_tmp"
    try:
        if sys.platform.startswith("linux") and scratch_type == "memory":
            await destroy_scratch_filesystem(scratch_dir)
            await destroy_scratch_filesystem(tmp_dir)
            await loop.run_in_executor(None, shutil.rmtree, scratch_dir)
            await loop.run_in_executor(None, shutil.rmtree, tmp_dir)
        elif sys.platform.startswith("linux") and scratch_type == "hostfile":
            await destroy_loop_filesystem(scratch_root, kernel_id)
        else:
            await loop.run_in_executor(None, shutil.rmtree, scratch_dir)
    except CalledProcessError:
        pass
    except FileNotFoundError:
        pass


def _DockerError_reduce(self):
    return (
        type(self),
        (self.status, {"message": self.message}, *self.args),
    )


def _DockerContainerError_reduce(self):
    return (
        type(self),
        (self.status, {"message": self.message}, self.container_id, *self.args),
    )


@dataclass
class DockerPurgeImageReq:
    image: str
    force: bool
    noprune: bool


class DockerKernelCreationContext(AbstractKernelCreationContext[DockerKernel]):
    scratch_dir: Path
    tmp_dir: Path
    config_dir: Path
    work_dir: Path
    container_configs: List[Mapping[str, Any]]
    domain_socket_proxies: List[DomainSocketProxy]
    computer_docker_args: Dict[str, Any]
    port_pool: Set[int]
    agent_sockpath: Path
    resource_lock: asyncio.Lock
    cluster_ssh_port_mapping: Optional[ClusterSSHPortMapping]
    gwbridge_subnet: Optional[str]

    network_plugin_ctx: NetworkPluginContext

    def __init__(
        self,
        ownership_data: KernelOwnershipData,
        event_producer: EventProducer,
        kernel_image: ImageRef,
        kernel_config: KernelCreationConfig,
        distro: str,
        local_config: AgentUnifiedConfig,
        computers: MutableMapping[DeviceName, ComputerContext],
        port_pool: Set[int],
        agent_sockpath: Path,
        resource_lock: asyncio.Lock,
        network_plugin_ctx: NetworkPluginContext,
        restarting: bool = False,
        cluster_ssh_port_mapping: Optional[ClusterSSHPortMapping] = None,
        gwbridge_subnet: Optional[str] = None,
    ) -> None:
        super().__init__(
            ownership_data,
            event_producer,
            kernel_image,
            kernel_config,
            distro,
            local_config,
            computers,
            restarting=restarting,
        )
        kernel_id = ownership_data.kernel_id
        scratch_dir = (self.local_config.container.scratch_root / str(kernel_id)).resolve()
        tmp_dir = (self.local_config.container.scratch_root / f"{kernel_id}_tmp").resolve()

        self.scratch_dir = scratch_dir
        self.tmp_dir = tmp_dir
        self.config_dir = scratch_dir / "config"
        self.work_dir = scratch_dir / "work"
        self.uid = kernel_config["uid"]
        self.main_gid = kernel_config["main_gid"]
        self.supplementary_gids = set(kernel_config["supplementary_gids"])

        self.port_pool = port_pool
        self.agent_sockpath = agent_sockpath
        self.resource_lock = resource_lock

        self.container_configs = []
        self.domain_socket_proxies = []
        self.computer_docker_args = {}

        self.cluster_ssh_port_mapping = cluster_ssh_port_mapping
        self.gwbridge_subnet = gwbridge_subnet

        self.network_plugin_ctx = network_plugin_ctx

    @override
    def get_overriding_uid(self) -> Optional[int]:
        return self.uid

    @override
    def get_overriding_gid(self) -> Optional[int]:
        return self.main_gid

    @override
    def get_supplementary_gids(self) -> set[int]:
        return self.supplementary_gids

    def _kernel_resource_spec_read(self, filename):
        with open(filename, "r") as f:
            resource_spec = KernelResourceSpec.read_from_file(f)
        return resource_spec

    async def get_extra_envs(self) -> Mapping[str, str]:
        return {}

    async def prepare_resource_spec(self) -> Tuple[KernelResourceSpec, Optional[Mapping[str, Any]]]:
        loop = current_loop()
        if self.restarting:
            resource_spec = await loop.run_in_executor(
                None, self._kernel_resource_spec_read, self.config_dir / "resource.txt"
            )
            resource_opts = None
        else:
            slots = ResourceSlot.from_json(self.kernel_config["resource_slots"])
            # Ensure that we have intrinsic slots.
            assert SlotName("cpu") in slots
            assert SlotName("mem") in slots
            # accept unknown slot type with zero values
            # but reject if they have non-zero values.
            for st, sv in slots.items():
                if st not in known_slot_types and sv != Decimal(0):
                    raise UnsupportedResource(st)
            # sanitize the slots
            current_resource_slots.set(known_slot_types)
            slots = slots.normalize_slots(ignore_unknown=True)
            resource_spec = KernelResourceSpec(
                allocations={},
                slots={**slots},  # copy
                mounts=[],
                scratch_disk_size=0,  # TODO: implement (#70)
            )
            resource_opts = self.kernel_config.get("resource_opts", {})
        return resource_spec, resource_opts

    def _chown_paths_if_root(
        self, paths: Iterable[Path], uid: Optional[int], gid: Optional[int]
    ) -> None:
        if os.geteuid() == 0:  # only possible when I am root.
            for p in paths:
                if KernelFeatures.UID_MATCH in self.kernel_features:
                    valid_uid = uid if uid is not None else self.local_config.container.kernel_uid
                    valid_gid = gid if gid is not None else self.local_config.container.kernel_gid
                else:
                    stat = os.stat(p)
                    valid_uid = uid if uid is not None else stat.st_uid
                    valid_gid = gid if gid is not None else stat.st_gid
                try:
                    int_uid = int(valid_uid)
                    int_gid = int(valid_gid)
                except (TypeError, ValueError):
                    log.exception(
                        "invalid uid/gid to chown: {}/{}, skip chown", valid_uid, valid_gid
                    )
                    continue
                try:
                    os.chown(p, int_uid, int_gid)
                except OSError as e:
                    log.exception(
                        "failed to chown {} to {}/{} (error: {})", p, int_uid, int_gid, repr(e)
                    )

    async def prepare_scratch(self) -> None:
        loop = current_loop()

        # Create the scratch, config, and work directories.
        scratch_type = self.local_config.container.scratch_type
        scratch_root = self.local_config.container.scratch_root
        scratch_size = self.local_config.container.scratch_size

        if sys.platform.startswith("linux") and scratch_type == "memory":
            await loop.run_in_executor(None, partial(self.tmp_dir.mkdir, exist_ok=True))
            await create_scratch_filesystem(self.scratch_dir, 64)
            await create_scratch_filesystem(self.tmp_dir, 64)
        elif sys.platform.startswith("linux") and scratch_type == "hostfile":
            await create_loop_filesystem(scratch_root, scratch_size, self.kernel_id)
        else:
            await loop.run_in_executor(None, partial(self.scratch_dir.mkdir, exist_ok=True))

        def _create_scratch_dirs():
            self.config_dir.mkdir(parents=True, exist_ok=True)
            self.config_dir.chmod(0o755)
            self.work_dir.mkdir(parents=True, exist_ok=True)
            self.work_dir.chmod(0o755)

        await loop.run_in_executor(None, _create_scratch_dirs)

        if not self.restarting:
            # Since these files are bind-mounted inside a bind-mounted directory,
            # we need to touch them first to avoid their "ghost" files are created
            # as root in the host-side filesystem, which prevents deletion of scratch
            # directories when the agent is running as non-root.
            def _clone_dotfiles():
                jupyter_custom_css_path = Path(
                    pkg_resources.resource_filename("ai.backend.runner", "jupyter-custom.css")
                )
                logo_path = Path(pkg_resources.resource_filename("ai.backend.runner", "logo.svg"))
                font_path = Path(pkg_resources.resource_filename("ai.backend.runner", "roboto.ttf"))
                font_italic_path = Path(
                    pkg_resources.resource_filename("ai.backend.runner", "roboto-italic.ttf")
                )
                bashrc_path = Path(pkg_resources.resource_filename("ai.backend.runner", ".bashrc"))
                bash_profile_path = Path(
                    pkg_resources.resource_filename("ai.backend.runner", ".bash_profile")
                )
                zshrc_path = Path(pkg_resources.resource_filename("ai.backend.runner", ".zshrc"))
                vimrc_path = Path(pkg_resources.resource_filename("ai.backend.runner", ".vimrc"))
                tmux_conf_path = Path(
                    pkg_resources.resource_filename("ai.backend.runner", ".tmux.conf")
                )
                jupyter_custom_dir = self.work_dir / ".jupyter" / "custom"
                jupyter_custom_dir.mkdir(parents=True, exist_ok=True)
                shutil.copy(jupyter_custom_css_path.resolve(), jupyter_custom_dir / "custom.css")
                shutil.copy(logo_path.resolve(), jupyter_custom_dir / "logo.svg")
                shutil.copy(font_path.resolve(), jupyter_custom_dir / "roboto.ttf")
                shutil.copy(font_italic_path.resolve(), jupyter_custom_dir / "roboto-italic.ttf")
                shutil.copy(bashrc_path.resolve(), self.work_dir / ".bashrc")
                shutil.copy(bash_profile_path.resolve(), self.work_dir / ".bash_profile")
                shutil.copy(zshrc_path.resolve(), self.work_dir / ".zshrc")
                shutil.copy(vimrc_path.resolve(), self.work_dir / ".vimrc")
                shutil.copy(tmux_conf_path.resolve(), self.work_dir / ".tmux.conf")

                def chown_scratch(uid: Optional[int], gid: Optional[int]) -> None:
                    paths = [
                        self.work_dir,
                        self.work_dir / ".jupyter",
                        self.work_dir / ".jupyter" / "custom",
                        self.work_dir / ".bashrc",
                        self.work_dir / ".bash_profile",
                        self.work_dir / ".zshrc",
                        self.work_dir / ".vimrc",
                        self.work_dir / ".tmux.conf",
                    ]
                    self._chown_paths_if_root(paths, uid, gid)

                do_override = False
                if (ouid := self.get_overriding_uid()) is not None:
                    do_override = True

                if (ogid := self.get_overriding_gid()) is not None:
                    do_override = True

                if do_override:
                    chown_scratch(ouid, ogid)
                else:
                    if KernelFeatures.UID_MATCH in self.kernel_features:
                        chown_scratch(
                            self.local_config.container.kernel_uid,
                            self.local_config.container.kernel_gid,
                        )

            await loop.run_in_executor(None, _clone_dotfiles)

    async def get_intrinsic_mounts(self) -> Sequence[Mount]:
        loop = current_loop()

        # scratch/config/tmp mounts
        mounts: List[Mount] = [
            Mount(
                MountTypes.BIND, self.config_dir, Path("/home/config"), MountPermission.READ_ONLY
            ),
            Mount(MountTypes.BIND, self.work_dir, Path("/home/work"), MountPermission.READ_WRITE),
        ]
        if (
            sys.platform.startswith("linux")
            and self.local_config.container.scratch_type == ScratchType.MEMORY
        ):
            mounts.append(
                Mount(
                    MountTypes.BIND,
                    self.tmp_dir,
                    Path("/tmp"),
                    MountPermission.READ_WRITE,
                )
            )
        # /etc/localtime and /etc/timezone mounts
        if sys.platform.startswith("linux"):
            localtime_file = Path("/etc/localtime")
            timezone_file = Path("/etc/timezone")
            if localtime_file.exists():
                mounts.append(
                    Mount(
                        type=MountTypes.BIND,
                        source=localtime_file,
                        target=localtime_file,
                        permission=MountPermission.READ_ONLY,
                    )
                )
            if timezone_file.exists():
                mounts.append(
                    Mount(
                        type=MountTypes.BIND,
                        source=timezone_file,
                        target=timezone_file,
                        permission=MountPermission.READ_ONLY,
                    )
                )
        # lxcfs mounts
        lxcfs_root = Path("/var/lib/lxcfs")
        if lxcfs_root.is_dir():
            mounts.extend(
                Mount(
                    MountTypes.BIND,
                    lxcfs_proc_path,
                    "/" / lxcfs_proc_path.relative_to(lxcfs_root),
                    MountPermission.READ_WRITE,
                )
                for lxcfs_proc_path in (lxcfs_root / "proc").iterdir()
                if lxcfs_proc_path.stat().st_size > 0
            )
            mounts.extend(
                Mount(
                    MountTypes.BIND,
                    lxcfs_root / path,
                    "/" / Path(path),
                    MountPermission.READ_WRITE,
                )
                for path in [
                    "sys/devices/system/cpu",
                    "sys/devices/system/cpu/online",
                ]
                if Path(lxcfs_root / path).exists()
            )

        # extra mounts
        async with closing_async(Docker()) as docker:
            extra_mount_list = await get_extra_volumes(docker, self.image_ref.short)
        mounts.extend(
            Mount(MountTypes.VOLUME, v.name, v.container_path, v.mode) for v in extra_mount_list
        )

        # debug mounts
        if self.local_config.debug.coredump.enabled:
            mounts.append(
                Mount(
                    MountTypes.BIND,
                    self.local_config.debug.coredump.path,
                    self.local_config.debug.coredump.core_path,
                    MountPermission.READ_WRITE,
                )
            )

        # agent-socket mount
        if sys.platform != "darwin":
            mounts.append(
                Mount(
                    MountTypes.BIND,
                    self.agent_sockpath,
                    Path("/opt/kernel/agent.sock"),
                    MountPermission.READ_WRITE,
                )
            )
        ipc_base_path = self.local_config.agent.ipc_base_path

        # domain-socket proxy mount
        # (used for special service containers such image importer)
        for host_sock_path in self.internal_data.get("domain_socket_proxies", []):
            await loop.run_in_executor(
                None, partial((ipc_base_path / "proxy").mkdir, parents=True, exist_ok=True)
            )
            host_proxy_path = ipc_base_path / "proxy" / f"{secrets.token_hex(12)}.sock"
            proxy_server = await asyncio.start_unix_server(
                aiotools.apartial(proxy_connection, host_sock_path), str(host_proxy_path)
            )
            await loop.run_in_executor(None, host_proxy_path.chmod, 0o666)
            self.domain_socket_proxies.append(
                DomainSocketProxy(
                    Path(host_sock_path),
                    host_proxy_path,
                    proxy_server,
                )
            )
            mounts.append(
                Mount(
                    MountTypes.BIND,
                    host_proxy_path,
                    host_sock_path,
                    MountPermission.READ_WRITE,
                )
            )

        return mounts

    def resolve_krunner_filepath(self, filename) -> Path:
        return Path(
            pkg_resources.resource_filename(
                "ai.backend.runner",
                "../" + filename,
            )
        ).resolve()

    def get_runner_mount(
        self,
        type: MountTypes,
        src: Union[str, Path],
        target: Union[str, Path],
        perm: MountPermission = MountPermission.READ_ONLY,
        opts: Optional[Mapping[str, Any]] = None,
    ) -> Mount:
        return Mount(
            type,
            Path(src),
            Path(target),
            MountPermission(perm),
            opts=opts,
        )

    async def apply_network(self, cluster_info: ClusterInfo) -> None:
        # FIXME: find out way to inect network ID to kernel resource spec
        match cluster_info["network_config"].get("mode"):
            case "bridge":
                self.container_configs.append({
                    "HostConfig": {
                        "NetworkMode": cluster_info["network_config"]["network_name"],
                    },
                    "NetworkingConfig": {
                        "EndpointsConfig": {
                            cluster_info["network_config"]["network_name"]: {
                                "Aliases": [self.kernel_config["cluster_hostname"]],
                            },
                        },
                    },
                })
            case mode if mode:
                try:
                    plugin = self.network_plugin_ctx.plugins[mode]
                except KeyError:
                    raise RuntimeError(f"Network plugin {mode} not loaded!")

                container_config = await plugin.join_network(
                    self.kernel_config, cluster_info, **cluster_info["network_config"]
                )
                self.container_configs.append(container_config)
                if self.gwbridge_subnet is not None:
                    self.container_configs.append({
                        "Env": [f"OMPI_MCA_btl_tcp_if_exclude=127.0.0.1/32,{self.gwbridge_subnet}"],
                    })
        if self.local_config.container.alternative_bridge is not None:
            self.container_configs.append({
                "HostConfig": {
                    "NetworkMode": self.local_config.container.alternative_bridge,
                },
            })
        # RDMA mounts
        ib_root = Path("/dev/infiniband")
        if ib_root.is_dir() and (ib_root / "uverbs0").exists():
            self.container_configs.append({
                "HostConfig": {
                    "Devices": [
                        {
                            "PathOnHost": "/dev/infiniband",
                            "PathInContainer": "/dev/infiniband",
                            "CgroupPermissions": "rwm",
                        },
                    ],
                },
            })

    async def prepare_ssh(self, cluster_info: ClusterInfo) -> None:
        sshkey = cluster_info["ssh_keypair"]
        if sshkey is None:
            return

        def _write_config():
            try:
                priv_key_path = self.config_dir / "ssh" / "id_cluster"
                pub_key_path = self.config_dir / "ssh" / "id_cluster.pub"
                priv_key_path.parent.mkdir(parents=True, exist_ok=True)
                priv_key_path.write_text(sshkey["private_key"])
                pub_key_path.write_text(sshkey["public_key"])

                def chown_keyfile(uid: Optional[int], gid: Optional[int]) -> None:
                    self._chown_paths_if_root([priv_key_path, pub_key_path], uid, gid)

                do_override = False
                if (ouid := self.get_overriding_uid()) is not None:
                    do_override = True

                if (ogid := self.get_overriding_gid()) is not None:
                    do_override = True

                if do_override:
                    chown_keyfile(ouid, ogid)
                else:
                    if KernelFeatures.UID_MATCH in self.kernel_features:
                        chown_keyfile(
                            self.local_config.container.kernel_uid,
                            self.local_config.container.kernel_gid,
                        )
                priv_key_path.chmod(0o600)
                if cluster_ssh_port_mapping := cluster_info["cluster_ssh_port_mapping"]:
                    port_mapping_json_path = self.config_dir / "ssh" / "port-mapping.json"
                    port_mapping_json_path.write_bytes(dump_json(cluster_ssh_port_mapping))
            except Exception:
                log.exception("error while writing cluster keypair")

        current_loop().run_in_executor(None, _write_config)  # ???

    async def process_mounts(self, mounts: Sequence[Mount]):
        def fix_unsupported_perm(folder_perm: MountPermission) -> MountPermission:
            if folder_perm == MountPermission.RW_DELETE:
                # TODO: enforce readable/writable but not deletable
                # (Currently docker's READ_WRITE includes DELETE)
                return MountPermission.READ_WRITE
            return folder_perm

        container_config = {
            "HostConfig": {
                "Mounts": [
                    {
                        "Target": str(mount.target),
                        "Source": str(mount.source),
                        "Type": mount.type.value,
                        "ReadOnly": (
                            fix_unsupported_perm(mount.permission) == MountPermission.READ_ONLY
                        ),
                        f"{mount.type.value.capitalize()}Options": mount.opts if mount.opts else {},
                    }
                    for mount in mounts
                ],
            },
        }
        self.container_configs.append(container_config)

    async def apply_accelerator_allocation(
        self,
        computer: AbstractComputePlugin,
        device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> None:
        async with closing_async(Docker()) as docker:
            update_nested_dict(
                self.computer_docker_args,
                await computer.generate_docker_args(docker, device_alloc),
            )

    async def generate_accelerator_mounts(
        self,
        computer: AbstractComputePlugin,
        device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> List[MountInfo]:
        src_path = self.config_dir / str(computer.key)
        src_path.mkdir(exist_ok=True)
        return await computer.generate_mounts(src_path, device_alloc)

    async def prepare_container(
        self,
        resource_spec: KernelResourceSpec,
        environ: Mapping[str, str],
        service_ports: List[ServicePort],
        cluster_info: ClusterInfo,
    ) -> DockerKernel:
        loop = current_loop()
        ouid = self.get_overriding_uid()
        ogid = self.get_overriding_gid()

        if self.restarting:
            pass
        else:
            # Create bootstrap.sh into workdir if needed
            if bootstrap := self.kernel_config.get("bootstrap_script"):

                def _write_user_bootstrap_script():
                    (self.work_dir / "bootstrap.sh").write_text(bootstrap)
                    if ouid is not None or ogid is not None:
                        self._chown_paths_if_root([self.work_dir / "bootstrap.sh"], ouid, ogid)
                    else:
                        if KernelFeatures.UID_MATCH in self.kernel_features:
                            self._chown_paths_if_root(
                                [self.work_dir / "bootstrap.sh"],
                                self.local_config.container.kernel_uid,
                                self.local_config.container.kernel_gid,
                            )

                await loop.run_in_executor(None, _write_user_bootstrap_script)

            with StringIO() as buf:
                for k, v in environ.items():
                    buf.write(f"{k}={v}\n")
                accel_envs = self.computer_docker_args.get("Env", [])
                for env in accel_envs:
                    buf.write(f"{env}\n")
                await loop.run_in_executor(
                    None,
                    (self.config_dir / "environ.txt").write_bytes,
                    buf.getvalue().encode("utf8"),
                )

            with StringIO() as buf:
                resource_spec.write_to_file(buf)
                for dev_type, device_alloc in resource_spec.allocations.items():
                    device_plugin = self.computers[dev_type].instance
                    kvpairs = await device_plugin.generate_resource_data(device_alloc)
                    for k, v in kvpairs.items():
                        buf.write(f"{k}={v}\n")
                await loop.run_in_executor(
                    None,
                    (self.config_dir / "resource.txt").write_bytes,
                    buf.getvalue().encode("utf8"),
                )

        shutil.copyfile(self.config_dir / "environ.txt", self.config_dir / "environ_base.txt")
        shutil.copyfile(self.config_dir / "resource.txt", self.config_dir / "resource_base.txt")

        # TODO: refactor out dotfiles/sshkey initialization to the base agent?
        docker_creds = self.internal_data.get("docker_credentials")
        if docker_creds:
            await loop.run_in_executor(
                None,
                (self.config_dir / "docker-creds.json").write_bytes,
                dump_json(docker_creds),
            )

        # Create SSH keypair only if ssh_keypair internal_data exists and
        # /home/work/.ssh folder is not mounted.
        if self.internal_data.get("ssh_keypair"):
            for mount in resource_spec.mounts:
                container_path = str(mount).split(":")[1]
                if container_path == "/home/work/.ssh":
                    break
            else:
                pubkey = self.internal_data["ssh_keypair"]["public_key"].encode("ascii")
                privkey = self.internal_data["ssh_keypair"]["private_key"].encode("ascii")
                ssh_dir = self.work_dir / ".ssh"

                def _populate_ssh_config():
                    ssh_dir.mkdir(parents=True, exist_ok=True)
                    ssh_dir.chmod(0o700)
                    (ssh_dir / "authorized_keys").write_bytes(pubkey)
                    (ssh_dir / "authorized_keys").chmod(0o600)
                    if not (ssh_dir / "id_rsa").is_file():
                        (ssh_dir / "id_rsa").write_bytes(privkey)
                        (ssh_dir / "id_rsa").chmod(0o600)
                    (self.work_dir / "id_container").write_bytes(privkey)
                    (self.work_dir / "id_container").chmod(0o600)

                    def chown_idfile(uid: Optional[int], gid: Optional[int]) -> None:
                        paths = [
                            ssh_dir,
                            ssh_dir / "authorized_keys",
                            ssh_dir / "id_rsa",
                            self.work_dir / "id_container",
                        ]
                        self._chown_paths_if_root(paths, uid, gid)

                    if ouid is not None or ogid is not None:
                        chown_idfile(ouid, ogid)
                    else:
                        if KernelFeatures.UID_MATCH in self.kernel_features:
                            chown_idfile(
                                self.local_config.container.kernel_uid,
                                self.local_config.container.kernel_gid,
                            )

                await loop.run_in_executor(None, _populate_ssh_config)

        # higher priority dotfiles are stored last to support overwriting
        for dotfile in self.internal_data.get("dotfiles", []):
            if dotfile["path"].startswith("/"):
                if dotfile["path"].startswith("/home/"):
                    path_arr = dotfile["path"].split("/")
                    file_path: Path = self.scratch_dir / "/".join(path_arr[2:])
                else:
                    file_path = Path(dotfile["path"])
            else:
                file_path = self.work_dir / dotfile["path"]
            file_path.parent.mkdir(parents=True, exist_ok=True)

            dotfile_content = dotfile["data"]
            if not dotfile_content.endswith("\n"):
                dotfile_content += "\n"
            await loop.run_in_executor(None, file_path.write_text, dotfile_content)

            tmp = Path(file_path)
            tmp_paths: list[Path] = []
            while tmp != self.work_dir:
                tmp.chmod(int(dotfile["perm"], 8))
                tmp_paths.append(tmp)
                tmp = tmp.parent
            if ouid is not None or ogid is not None:
                self._chown_paths_if_root(tmp_paths, ouid, ogid)
            else:
                if KernelFeatures.UID_MATCH in self.kernel_features:
                    self._chown_paths_if_root(
                        tmp_paths,
                        self.local_config.container.kernel_uid,
                        self.local_config.container.kernel_gid,
                    )

        kernel_obj = DockerKernel(
            self.ownership_data,
            self.kernel_config["network_id"],
            self.image_ref,
            self.kspec_version,
            cluster_info["network_config"].get("mode", "bridge"),
            agent_config=self.local_config.model_dump(by_alias=True),
            service_ports=service_ports,
            resource_spec=resource_spec,
            environ=environ,
            data={},
        )
        return kernel_obj

    @property
    @override
    def repl_ports(self) -> Sequence[int]:
        return (2000, 2001)

    @property
    @override
    def protected_services(self) -> Sequence[str]:
        rgtype: ResourceGroupType = self.local_config.agent.scaling_group_type
        match rgtype:
            case ResourceGroupType.COMPUTE:
                return ()
            case ResourceGroupType.STORAGE:
                return ("ttyd",)

    async def _apply_seccomp_profile(self, container_config: MutableMapping[str, Any]) -> None:
        default_seccomp_path = self.resolve_krunner_filepath("runner/default-seccomp.json")

        if not default_seccomp_path.exists():
            log.warning(
                "Default seccomp profile file not found in the expected path! Skipped the application of additional syscalls."
            )
            return

        async with aiofiles.open(default_seccomp_path, mode="r") as fp:
            seccomp_profile = load_json(await fp.read())

            additional_allowed_syscalls = self.additional_allowed_syscalls
            additional_allowed_syscall_rule = {
                "names": additional_allowed_syscalls,
                "action": "SCMP_ACT_ALLOW",
                "args": [],
                "comment": "Additionally allowed syscalls by Backend.AI Agent",
            }
            seccomp_profile["syscalls"].append(additional_allowed_syscall_rule)

            container_config["HostConfig"]["SecurityOpt"] = [
                f"seccomp={json.dumps(seccomp_profile)}"
            ]

    async def start_container(
        self,
        kernel_obj: AbstractKernel,
        cmdargs: List[str],
        resource_opts,
        preopen_ports,
        cluster_info: ClusterInfo,
    ) -> Mapping[str, Any]:
        loop = current_loop()
        resource_spec = kernel_obj.resource_spec
        service_ports = kernel_obj.service_ports
        environ = kernel_obj.environ
        image_labels = self.kernel_config["image"]["labels"]

        # PHASE 4: Run!
        container_bind_host = self.local_config.container.bind_host
        advertised_kernel_host = self.local_config.container.advertised_host
        if len(service_ports) + len(self.repl_ports) > len(self.port_pool):
            raise RuntimeError(
                f"Container ports are not sufficiently available. (remaining ports: {self.port_pool})"
            )
        exposed_ports = [*self.repl_ports]
        host_ports = [self.port_pool.pop() for _ in self.repl_ports]
        host_ips = []
        for sport in service_ports:
            exposed_ports.extend(sport["container_ports"])
            if (
                sport["name"] == "sshd"
                and self.cluster_ssh_port_mapping
                and (
                    ssh_host_port := self.cluster_ssh_port_mapping.get(
                        self.kernel_config["cluster_hostname"]
                    )
                )
            ):
                host_ports.append(ssh_host_port[1])
            else:
                hport = self.port_pool.pop()
                host_ports.append(hport)
        protected_service_ports: set[int] = set()
        for sport in service_ports:
            if sport["name"] in self.protected_services:
                protected_service_ports.update(sport["container_ports"])
        for eport in exposed_ports:
            if eport in self.repl_ports:  # always protected
                host_ips.append("127.0.0.1")
            elif eport in protected_service_ports:  # check if protected by resource group type
                host_ips.append("127.0.0.1")
            else:
                host_ips.append(str(container_bind_host))
        assert len(host_ips) == len(host_ports) == len(exposed_ports)

        if (mode := cluster_info["network_config"].get("mode")) and mode != "bridge":
            try:
                plugin = self.network_plugin_ctx.plugins[mode]
            except KeyError:
                raise RuntimeError(f"Network plugin {mode} not loaded!")
            if ContainerNetworkCapability.GLOBAL in (await plugin.get_capabilities()):
                await plugin.prepare_port_forward(
                    kernel_obj,
                    str(container_bind_host),
                    [
                        (host_port, container_port)
                        for host_port, container_port in zip(host_ports, exposed_ports)
                    ],
                )

        container_log_size = self.local_config.container_logs.max_length
        container_log_file_count = 5
        container_log_file_size = BinarySize(container_log_size // container_log_file_count)

        if self.image_ref.is_local:
            image = self.image_ref.short
        else:
            image = self.image_ref.canonical

        container_config: MutableMapping[str, Any] = {
            "Image": image,
            "Tty": True,
            "OpenStdin": True,
            "Privileged": False,
            "StopSignal": "SIGINT",
            "ExposedPorts": {f"{port}/tcp": {} for port in exposed_ports},
            "EntryPoint": ["/opt/kernel/entrypoint.sh"],
            "Cmd": cmdargs,
            "Env": [f"{k}={v}" for k, v in environ.items()],
            "WorkingDir": "/home/work",
            "Hostname": self.kernel_config["cluster_hostname"],
            "Labels": {
                LabelName.KERNEL_ID: str(self.kernel_id),
                LabelName.SESSION_ID: str(self.session_id),
                LabelName.OWNER_USER: self.ownership_data.owner_user_id_to_str,
                LabelName.OWNER_PROJECT: self.ownership_data.owner_project_id_to_str,
                LabelName.OWNER_AGENT: str(self.agent_id),
                LabelName.BLOCK_SERVICE_PORTS: (
                    "1" if self.internal_data.get("block_service_ports", False) else "0"
                ),
            },
            "HostConfig": {
                "Init": True,
                "PortBindings": {
                    f"{eport}/tcp": [{"HostPort": str(hport), "HostIp": hip}]
                    for eport, hport, hip in zip(exposed_ports, host_ports, host_ips)
                },
                "PublishAllPorts": False,  # we manage port mapping manually!
                "CapAdd": [
                    "IPC_LOCK",  # for hugepages and RDMA
                    "SYS_NICE",  # for NFS based GPUDirect Storage
                ],
                "Ulimits": [
                    get_safe_ulimit("nofile", 1048576, 1048576),
                    get_safe_ulimit("memlock", -1, -1),
                ],
                "LogConfig": {
                    "Type": "local",  # for efficient docker-specific storage
                    "Config": {
                        # these fields must be str
                        # (ref: https://docs.docker.com/config/containers/logging/local/)
                        "max-size": f"{container_log_file_size:s}",
                        "max-file": str(container_log_file_count),
                        "compress": "false",
                    },
                },
            },
        }

        await self._apply_seccomp_profile(container_config)

        # merge all container configs generated during prior preparation steps
        for c in self.container_configs:
            update_nested_dict(container_config, c)
        if self.local_config.container.sandbox_type == ContainerSandboxType.JAIL:
            update_nested_dict(
                container_config,
                {
                    "HostConfig": {
                        "SecurityOpt": ["seccomp=unconfined", "apparmor=unconfined"],
                        "CapAdd": ["SYS_PTRACE"],
                    },
                },
            )

        if resource_opts and resource_opts.get("shmem"):
            shmem = int(resource_opts.get("shmem", "0"))
            self.computer_docker_args["HostConfig"]["ShmSize"] = shmem
            self.computer_docker_args["HostConfig"]["MemorySwap"] -= shmem
            self.computer_docker_args["HostConfig"]["Memory"] -= shmem

        service_ports_label: list[str] = []
        service_ports_label += image_labels.get(LabelName.SERVICE_PORTS, "").split(",")
        service_ports_label += [f"{port_no}:preopen:{port_no}" for port_no in preopen_ports]

        container_config["Labels"][LabelName.SERVICE_PORTS] = ",".join([
            label for label in service_ports_label if label
        ])
        update_nested_dict(container_config, self.computer_docker_args)
        kernel_name = f"kernel.{self.image_ref.name.split('/')[-1]}.{self.kernel_id}"

        # optional local override of docker config
        extra_container_opts_name = "agent-docker-container-opts.json"
        for extra_container_opts_file in [
            Path("/etc/backend.ai") / extra_container_opts_name,
            Path.home() / ".config" / "backend.ai" / extra_container_opts_name,
            Path.cwd() / extra_container_opts_name,
        ]:
            if extra_container_opts_file.is_file():
                try:
                    extra_container_opts = load_json(extra_container_opts_file.read_bytes())
                    update_nested_dict(container_config, extra_container_opts)
                except IOError:
                    pass

        # The final container config is settled down here.
        if self.local_config.debug.log_kernel_config:
            log.debug("full container config: {!r}", pretty(container_config))

        async def _rollback_container_creation() -> None:
            await _clean_scratch(
                loop,
                self.local_config.container.scratch_type,
                self.local_config.container.scratch_root,
                self.kernel_id,
            )
            self.port_pool.update(host_ports)
            async with self.resource_lock:
                for dev_name, device_alloc in resource_spec.allocations.items():
                    self.computers[dev_name].alloc_map.free(device_alloc)

        # We are all set! Create and start the container.
        async with closing_async(Docker()) as docker:
            container: Optional[DockerContainer] = None
            try:
                container = await docker.containers.create(
                    config=container_config, name=kernel_name
                )
                assert container is not None
                cid = cast(str, container._id)
                async with AsyncFileWriter(
                    target_filename=self.config_dir / "resource.txt",
                    access_mode="a",
                ) as writer:
                    await writer.write(f"CID={cid}\n")

            except asyncio.CancelledError:
                if container is not None:
                    raise ContainerCreationError(
                        container_id=ContainerId(container.id),
                        message="Container creation cancelled",
                    )
                raise
            except Exception as e:
                # Oops, we have to restore the allocated resources!
                await _rollback_container_creation()
                if container is not None:
                    raise ContainerCreationError(
                        container_id=ContainerId(container.id),
                        message=f"unknown. {repr(e)}",
                    )
                raise

            try:
                await container.start()
            except asyncio.CancelledError:
                await _rollback_container_creation()
                raise ContainerCreationError(container_id=cid, message="Container start cancelled")
            except Exception as e:
                await _rollback_container_creation()
                raise ContainerCreationError(container_id=cid, message=f"unknown. {repr(e)}")

            if self.internal_data.get("sudo_session_enabled", False):
                exec = await container.exec(
                    [
                        # file ownership is guaranteed to be set as root:root since command is executed on behalf of root user
                        "sh",
                        "-c",
                        'mkdir -p /etc/sudoers.d && echo "work ALL=(ALL:ALL) NOPASSWD:ALL" > /etc/sudoers.d/01-bai-work',
                    ],
                    user="root",
                )
                shell_response = await exec.start(detach=True)
                if shell_response:
                    await _rollback_container_creation()
                    raise ContainerCreationError(
                        container_id=cid,
                        message=f"sudoers provision failed: {shell_response.decode()}",
                    )

            additional_network_names: Set[str] = set()
            for dev_name, device_alloc in resource_spec.allocations.items():
                n = await self.computers[dev_name].instance.get_docker_networks(device_alloc)
                additional_network_names |= set(n)

            for name in additional_network_names:
                network = await docker.networks.get(name)
                await network.connect({"Container": container._id})

            kernel_obj.container_id = container._id
            container_network_info: Optional[ContainerNetworkInfo] = None
            if (mode := cluster_info["network_config"].get("mode")) and mode != "bridge":
                try:
                    plugin = self.network_plugin_ctx.plugins[mode]
                except KeyError:
                    raise RuntimeError(f"Network plugin {mode} not loaded!")
                if ContainerNetworkCapability.GLOBAL in (await plugin.get_capabilities()):
                    container_network_info = await plugin.expose_ports(
                        kernel_obj,
                        str(container_bind_host),
                        [
                            (host_port, container_port)
                            for host_port, container_port in zip(host_ports, exposed_ports)
                        ],
                    )

            created_host_ports: Tuple[int, ...]
            repl_in_port = 0
            repl_out_port = 0
            if container_network_info:
                kernel_host = container_network_info.container_host
                port_map = container_network_info.services
                assert "replin" in port_map and "replout" in port_map

                repl_in_port = port_map["replin"][2000]
                repl_out_port = port_map["replout"][2001]
                stdin_port = 0  # left for legacy
                stdout_port = 0  # left for legacy

                for sport in service_ports:
                    created_host_ports = tuple(
                        port_map[sport["name"]][cport] for cport in sport["container_ports"]
                    )
                    sport["host_ports"] = created_host_ports
            else:
                kernel_host = advertised_kernel_host or container_bind_host
                ctnr_host_port_map: MutableMapping[int, int] = {}
                stdin_port = 0
                stdout_port = 0
                for idx, port in enumerate(exposed_ports):
                    ports: Optional[list[PortInfo]] = await container.port(port)
                    if not ports:
                        raise ContainerCreationError(
                            container_id=cid, message="Container port not found"
                        )
                    host_port = int(ports[0]["HostPort"])
                    if host_port != host_ports[idx]:
                        await _rollback_container_creation()
                        raise ContainerCreationError(
                            container_id=cid,
                            message=f"Port mapping mismatch. {host_port = }, {host_ports[idx] = }",
                        )
                    assert host_port == host_ports[idx]
                    if port == 2000:  # intrinsic
                        repl_in_port = host_port
                    elif port == 2001:  # intrinsic
                        repl_out_port = host_port
                    elif port == 2002:  # legacy
                        stdin_port = host_port
                    elif port == 2003:  # legacy
                        stdout_port = host_port
                    else:
                        ctnr_host_port_map[port] = host_port
                for sport in service_ports:
                    created_host_ports = tuple(
                        ctnr_host_port_map[cport] for cport in sport["container_ports"]
                    )
                    sport["host_ports"] = created_host_ports

        assert repl_in_port != 0, "repl_in_port should have been assigned."
        assert repl_out_port != 0, "repl_out_port should have been assigned."
        return {
            "container_id": container._id,
            "kernel_host": kernel_host,
            "repl_in_port": repl_in_port,
            "repl_out_port": repl_out_port,
            "stdin_port": stdin_port,  # legacy
            "stdout_port": stdout_port,  # legacy
            "host_ports": host_ports,
            "domain_socket_proxies": self.domain_socket_proxies,
            "block_service_ports": self.internal_data.get("block_service_ports", False),
        }


class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
    docker_info: Mapping[str, Any]
    monitor_docker_task: asyncio.Task
    agent_sockpath: Path
    agent_sock_task: asyncio.Task
    metadata_server: MetadataServer
    docker_ptask_group: aiotools.PersistentTaskGroup
    gwbridge_subnet: Optional[str]
    checked_invalid_images: Set[str]

    network_plugin_ctx: NetworkPluginContext

    def __init__(
        self,
        etcd: AsyncEtcd,
        local_config: AgentUnifiedConfig,
        *,
        stats_monitor: StatsPluginContext,
        error_monitor: ErrorPluginContext,
        skip_initial_scan: bool = False,
        agent_public_key: Optional[PublicKey],
    ) -> None:
        super().__init__(
            etcd,
            local_config,
            stats_monitor=stats_monitor,
            error_monitor=error_monitor,
            skip_initial_scan=skip_initial_scan,
            agent_public_key=agent_public_key,
        )
        self.checked_invalid_images = set()

    async def __ainit__(self) -> None:
        async with closing_async(Docker()) as docker:
            docker_host = ""
            match docker.connector:
                case aiohttp.TCPConnector():
                    assert docker.docker_host is not None
                    docker_host = docker.docker_host
                case aiohttp.NamedPipeConnector() | aiohttp.UnixConnector() as connector:
                    docker_host = connector.path
                case _:
                    docker_host = "(unknown)"
            log.info("accessing the local Docker daemon via {}", docker_host)
            docker_version = await docker.version()
            log.info(
                "running with Docker {0} with API {1}",
                docker_version["Version"],
                docker_version["ApiVersion"],
            )
            kernel_version = docker_version["KernelVersion"]
            if "linuxkit" in kernel_version:
                self.local_config.agent.docker_mode = "linuxkit"
            else:
                self.local_config.agent.docker_mode = "native"
            docker_info = await docker.system.info()
            docker_info = dict(docker_info)
            # Assume cgroup v1 if CgroupVersion key is absent
            if "CgroupVersion" not in docker_info:
                docker_info["CgroupVersion"] = "1"
            log.info(
                "Cgroup Driver: {0}, Cgroup Version: {1}",
                docker_info["CgroupDriver"],
                docker_info["CgroupVersion"],
            )
            self.docker_info = docker_info
        await super().__ainit__()
        try:
            async with Docker() as docker:
                gwbridge = await docker.networks.get("docker_gwbridge")
                gwbridge_info = await gwbridge.show()
                self.gwbridge_subnet = gwbridge_info["IPAM"]["Config"][0]["Subnet"]
        except (DockerError, KeyError, IndexError):
            self.gwbridge_subnet = None
        ipc_base_path = self.local_config.agent.ipc_base_path
        (ipc_base_path / "container").mkdir(parents=True, exist_ok=True)
        self.agent_sockpath = ipc_base_path / "container" / f"agent.{self.local_instance_id}.sock"
        # Workaround for Docker Desktop for Mac's UNIX socket mount failure with virtiofs
        if sys.platform != "darwin":
            socket_relay_name = f"backendai-socket-relay.{self.local_instance_id}"
            socket_relay_container = PersistentServiceContainer(
                "backendai-socket-relay:latest",
                {
                    "Cmd": [
                        f"UNIX-LISTEN:/ipc/{self.agent_sockpath.name},unlink-early,fork,mode=777",
                        f"TCP-CONNECT:127.0.0.1:{self.local_config.agent.agent_sock_port}",
                    ],
                    "HostConfig": {
                        "Mounts": [
                            {
                                "Type": "bind",
                                "Source": str(ipc_base_path / "container"),
                                "Target": "/ipc",
                            },
                        ],
                        "NetworkMode": "host",
                    },
                },
                name=socket_relay_name,
            )
            await socket_relay_container.ensure_running_latest()
        self.agent_sock_task = asyncio.create_task(self.handle_agent_socket())
        self.monitor_docker_task = asyncio.create_task(self.monitor_docker_events())
        self.docker_ptask_group = aiotools.PersistentTaskGroup()

        self.metadata_server = await MetadataServer.new(
            self.local_config,
            self.etcd,
            self.kernel_registry,
        )
        await self.metadata_server.start_server()
        # For legacy accelerator plugins
        self.docker = Docker()

        self.network_plugin_ctx = NetworkPluginContext(
            self.etcd, self.local_config.model_dump(by_alias=True)
        )
        await self.network_plugin_ctx.init(
            context=self,
            allowlist=self.local_config.agent.allow_network_plugins,
            blocklist=self.local_config.agent.block_network_plugins,
        )

    async def shutdown(self, stop_signal: signal.Signals):
        # Stop handling agent sock.
        if self.agent_sock_task is not None:
            self.agent_sock_task.cancel()
            await self.agent_sock_task
        if self.docker_ptask_group is not None:
            await self.docker_ptask_group.shutdown()

        try:
            await super().shutdown(stop_signal)
        finally:
            # Stop docker event monitoring.
            if self.monitor_docker_task is not None:
                self.monitor_docker_task.cancel()
                await self.monitor_docker_task

        await self.metadata_server.cleanup()
        if self.docker:
            await self.docker.close()

    @override
    def get_cgroup_path(self, controller: str, container_id: str) -> Path:
        driver = self.docker_info["CgroupDriver"]
        version = self.docker_info["CgroupVersion"]
        mount_point = get_cgroup_mount_point(version, controller)
        # See https://docs.docker.com/config/containers/runmetrics/#find-the-cgroup-for-a-given-container
        match driver:
            case "cgroupfs":
                cgroup = f"docker/{container_id}"
            case "systemd":
                cgroup = f"system.slice/docker-{container_id}.scope"
            case _:
                raise ValueError(f"Unsupported cgroup driver: {driver!r}")
        return mount_point / cgroup

    @override
    def get_cgroup_version(self) -> str:
        return self.docker_info["CgroupVersion"]

    async def load_resources(self) -> Mapping[DeviceName, AbstractComputePlugin]:
        return await load_resources(self.etcd, self.local_config.model_dump(by_alias=True))

    async def scan_available_resources(self) -> Mapping[SlotName, Decimal]:
        return await scan_available_resources(
            self.local_config.model_dump(by_alias=True),
            {name: cctx.instance for name, cctx in self.computers.items()},
        )

    async def extract_image_command(self, image: str) -> Optional[str]:
        async with closing_async(Docker()) as docker:
            result = await docker.images.get(image)
            return result["Config"].get("Cmd")

    async def enumerate_containers(
        self,
        status_filter: FrozenSet[ContainerStatus] = ACTIVE_STATUS_SET,
    ) -> Sequence[Tuple[KernelId, Container]]:
        result = []
        fetch_tasks = []
        async with closing_async(Docker()) as docker:
            for container in await docker.containers.list():

                async def _fetch_container_info(container):
                    kernel_id = "(unknown)"
                    try:
                        kernel_id = await get_kernel_id_from_container(container)
                        if kernel_id is None:
                            return
                        if container["State"]["Status"] in status_filter:
                            owner_id = AgentId(
                                container["Config"]["Labels"].get(LabelName.OWNER_AGENT, "")
                            )
                            if self.id == owner_id:
                                await container.show()
                                result.append(
                                    (
                                        kernel_id,
                                        container_from_docker_container(container),
                                    ),
                                )
                    except DockerError as e:
                        if e.status == HTTPStatus.NOT_FOUND:
                            log.warning(e.message)
                            return
                        raise
                    except asyncio.CancelledError:
                        pass
                    except Exception:
                        log.exception(
                            "error while fetching container information (cid:{}, k:{})",
                            container._id,
                            kernel_id,
                        )

                fetch_tasks.append(_fetch_container_info(container))

            await asyncio.gather(*fetch_tasks, return_exceptions=True)
        return result

    async def resolve_image_distro(self, image: ImageConfig) -> str:
        image_labels = image["labels"]
        distro = image_labels.get(LabelName.BASE_DISTRO)
        if distro:
            return distro

        async with Docker() as docker:
            image_id = image["digest"].partition(":")[-1]
            # check if distro data is available on redis cache
            cached_distro = await self.valkey_stat_client.get_image_distro(image_id)
            if cached_distro:
                return cached_distro

            container_config: dict[str, Any] = {
                "Image": image["canonical"],
                "Tty": True,
                "Privileged": False,
                "AttachStdin": False,
                "AttachStdout": True,
                "AttachStderr": True,
                "HostConfig": {
                    "Init": True,
                },
                "Entrypoint": [""],
                "Cmd": ["ldd", "--version"],
            }

            container = await docker.containers.create(container_config)
            await container.start()
            await container.wait()  # wait until container finishes to prevent race condition
            container_log = await container.log(stdout=True, stderr=True, follow=False)
            await container.stop()
            await container.delete()
            log.debug("response: {}", container_log)
            version_lines = container_log[0].splitlines()
            if m := LDD_GLIBC_REGEX.search(version_lines[0]):
                version = float(m.group(1))
                if version in known_glibc_distros:
                    distro = known_glibc_distros[version]
                else:
                    for idx, known_version in enumerate(known_glibc_distros.keys()):
                        if version < known_version:
                            distro = list(known_glibc_distros.values())[idx - 1]
                            break
                    else:
                        distro = list(known_glibc_distros.values())[-1]
            elif m := LDD_MUSL_REGEX.search(version_lines[0]):
                distro = "alpine3.8"
            else:
                raise RuntimeError("Could not determine the C library variant.")
            await self.valkey_stat_client.set_image_distro(image_id, distro)
            return distro

    async def scan_images(self) -> ScanImagesResult:
        async with closing_async(Docker()) as docker:
            all_images = await docker.images.list()
            scanned_images: dict[ImageCanonical, ScannedImage] = {}
            removed_images: dict[ImageCanonical, ScannedImage] = {}
            for image in all_images:
                if image["RepoTags"] is None:
                    continue
                for repo_tag in image["RepoTags"]:
                    if repo_tag.endswith("<none>"):
                        continue
                    try:
                        ImageRef.parse_image_str(repo_tag, "*")
                    except (InvalidImageName, InvalidImageTag) as e:
                        if repo_tag not in self.checked_invalid_images:
                            log.warning(
                                "Image name {} does not conform to Backend.AI's image naming rule. This image will be ignored. Details: {}",
                                repo_tag,
                                e,
                            )
                            self.checked_invalid_images.add(repo_tag)
                        continue

                    img_detail = await docker.images.inspect(repo_tag)
                    labels = (img_detail.get("Config") or {}).get("Labels")
                    if labels is None:
                        continue

                    kernelspec = int(labels.get(LabelName.KERNEL_SPEC, "1"))
                    if MIN_KERNELSPEC <= kernelspec <= MAX_KERNELSPEC:
                        scanned_images[ImageCanonical(repo_tag)] = ScannedImage(
                            canonical=repo_tag,
                            digest=img_detail["Id"],
                        )
            for added_image in scanned_images.keys() - self.images.keys():
                log.debug("found kernel image: {0}", added_image)

            for removed_image in self.images.keys() - scanned_images.keys():
                log.debug("removed kernel image: {0}", removed_image)
                removed_images[removed_image] = self.images[removed_image]

            return ScanImagesResult(
                scanned_images=scanned_images,
                removed_images=removed_images,
            )

    async def handle_agent_socket(self):
        """
        A simple request-reply socket handler for in-container processes.
        For ease of implementation in low-level languages such as C,
        it uses a simple C-friendly ZeroMQ-based multipart messaging protocol.

        The agent listens on a local TCP port and there is a socat relay
        that proxies this port via a UNIX domain socket mounted inside
        actual containers.  The reason for this is to avoid inode changes
        upon agent restarts by keeping the relay container running persistently,
        so that the mounted UNIX socket files don't get to refere a dangling pointer
        when the agent is restarted.

        Request message:
            The first part is the requested action as string,
            The second part and later are arguments.

        Reply message:
            The first part is a 32-bit integer (int in C)
                (0: success)
                (-1: generic unhandled error)
                (-2: invalid action)
            The second part and later are arguments.

        All strings are UTF-8 encoded.
        """
        terminating = False
        zmq_ctx = zmq.asyncio.Context()
        while True:
            agent_sock = zmq_ctx.socket(zmq.REP)
            try:
                agent_sock.bind(f"tcp://127.0.0.1:{self.local_config.agent.agent_sock_port}")
                while True:
                    msg = await agent_sock.recv_multipart()
                    if not msg:
                        break
                    try:
                        if msg[0] == b"host-pid-to-container-pid":
                            container_id = msg[1].decode()
                            host_pid = struct.unpack("i", msg[2])[0]
                            container_pid = await host_pid_to_container_pid(
                                container_id,
                                host_pid,
                            )
                            reply = [
                                struct.pack("i", 0),
                                struct.pack("i", container_pid),
                            ]
                        elif msg[0] == b"container-pid-to-host-pid":
                            container_id = msg[1].decode()
                            container_pid = struct.unpack("i", msg[2])[0]
                            host_pid = await container_pid_to_host_pid(container_id, container_pid)
                            reply = [
                                struct.pack("i", 0),
                                struct.pack("i", host_pid),
                            ]
                        elif msg[0] == b"is-jail-enabled":
                            reply = [
                                struct.pack("i", 0),
                                struct.pack(
                                    "i",
                                    (
                                        1
                                        if self.local_config.container.sandbox_type
                                        == ContainerSandboxType.JAIL
                                        else 0
                                    ),
                                ),
                            ]
                        else:
                            reply = [struct.pack("i", -2), b"Invalid action"]
                    except asyncio.CancelledError:
                        terminating = True
                        raise
                    except Exception as e:
                        log.exception("handle_agent_socket(): internal error")
                        reply = [struct.pack("i", -1), f"Error: {e}".encode("utf-8")]
                    await agent_sock.send_multipart(reply)
            except asyncio.CancelledError:
                terminating = True
                return
            except zmq.ZMQError:
                log.exception("handle_agent_socket(): zmq error")
                raise
            finally:
                agent_sock.close()
                if not terminating:
                    log.info("handle_agent_socket(): rebinding the socket")
                else:
                    zmq_ctx.destroy()

    async def push_image(
        self,
        image_ref: ImageRef,
        registry_conf: ImageRegistry,
        *,
        timeout: Optional[float] | Sentinel = Sentinel.TOKEN,
    ) -> None:
        if image_ref.is_local:
            return
        auth_config = None
        reg_user = registry_conf.get("username")
        reg_passwd = registry_conf.get("password")
        log.info("pushing image {} to registry", image_ref.canonical)
        if reg_user and reg_passwd:
            encoded_creds = base64.b64encode(f"{reg_user}:{reg_passwd}".encode("utf-8")).decode(
                "ascii"
            )
            auth_config = {
                "auth": encoded_creds,
            }

        async with closing_async(Docker()) as docker:
            kwargs: dict[str, Any] = {"auth": auth_config}
            if timeout != Sentinel.TOKEN:
                kwargs["timeout"] = timeout
            result = await docker.images.push(image_ref.canonical, **kwargs)

            if not result:
                raise RuntimeError("Failed to push image: unexpected return value from aiodocker")
            elif error := result[-1].get("error"):
                raise RuntimeError(f"Failed to push image: {error}")

    async def pull_image(
        self,
        image_ref: ImageRef,
        registry_conf: ImageRegistry,
        *,
        timeout: Optional[float],
    ) -> None:
        auth_config = None
        reg_user = registry_conf.get("username")
        reg_passwd = registry_conf.get("password")
        if reg_user and reg_passwd:
            encoded_creds = base64.b64encode(f"{reg_user}:{reg_passwd}".encode("utf-8")).decode(
                "ascii"
            )
            auth_config = {
                "auth": encoded_creds,
            }
        log.info("pulling image {} from registry", image_ref.canonical)
        async with closing_async(Docker()) as docker:
            result = await docker.images.pull(
                image_ref.canonical, auth=auth_config, timeout=timeout
            )

            if not result:
                raise RuntimeError("Failed to pull image: unexpected return value from aiodocker")
            elif error := result[-1].get("error"):
                raise RuntimeError(f"Failed to pull image: {error}")

    async def _purge_image(self, docker: Docker, request: DockerPurgeImageReq) -> PurgeImageResp:
        try:
            await docker.images.delete(request.image, force=request.force, noprune=request.noprune)
            return PurgeImageResp.success(image=request.image)
        except Exception as e:
            log.error(f'Failed to purge image "{request.image}": {e}')
            return PurgeImageResp.failure(image=request.image, error=str(e))

    async def purge_images(self, request: PurgeImagesReq) -> PurgeImagesResp:
        async with closing_async(Docker()) as docker:
            async with TaskGroup() as tg:
                tasks = [
                    tg.create_task(
                        self._purge_image(
                            docker,
                            DockerPurgeImageReq(
                                image=image, force=request.force, noprune=request.noprune
                            ),
                        )
                    )
                    for image in request.images
                ]

        results = []
        for task in tasks:
            deleted_info = task.result()
            results.append(deleted_info)

        return PurgeImagesResp(responses=results)

    async def check_image(
        self, image_ref: ImageRef, image_id: str, auto_pull: AutoPullBehavior
    ) -> bool:
        try:
            async with closing_async(Docker()) as docker:
                image_info = await docker.images.inspect(image_ref.canonical)
                if auto_pull == AutoPullBehavior.DIGEST:
                    if image_info["Id"] != image_id:
                        return True
            log.info("found the local up-to-date image for {}", image_ref.canonical)
        except DockerError as e:
            if e.status == HTTPStatus.NOT_FOUND:
                if auto_pull == AutoPullBehavior.DIGEST:
                    return True
                elif auto_pull == AutoPullBehavior.TAG:
                    return True
                elif auto_pull == AutoPullBehavior.NONE:
                    raise ImageNotAvailable(image_ref)
            else:
                raise
        return False

    async def init_kernel_context(
        self,
        ownership_data: KernelOwnershipData,
        kernel_image: ImageRef,
        kernel_config: KernelCreationConfig,
        *,
        restarting: bool = False,
        cluster_ssh_port_mapping: Optional[ClusterSSHPortMapping] = None,
    ) -> DockerKernelCreationContext:
        distro = await self.resolve_image_distro(kernel_config["image"])
        return DockerKernelCreationContext(
            ownership_data,
            self.event_producer,
            kernel_image,
            kernel_config,
            distro,
            self.local_config,
            self.computers,
            self.port_pool,
            self.agent_sockpath,
            self.resource_lock,
            self.network_plugin_ctx,
            restarting=restarting,
            cluster_ssh_port_mapping=cluster_ssh_port_mapping,
            gwbridge_subnet=self.gwbridge_subnet,
        )

    async def restart_kernel__load_config(
        self,
        kernel_id: KernelId,
        name: str,
    ) -> bytes:
        loop = current_loop()
        scratch_dir = (self.local_config.container.scratch_root / str(kernel_id)).resolve()
        config_dir = scratch_dir / "config"
        return await loop.run_in_executor(
            None,
            (config_dir / name).read_bytes,
        )

    async def restart_kernel__store_config(
        self,
        kernel_id: KernelId,
        name: str,
        data: bytes,
    ) -> None:
        loop = current_loop()
        scratch_dir = (self.local_config.container.scratch_root / str(kernel_id)).resolve()
        config_dir = scratch_dir / "config"

        def _write_bytes(data: bytes) -> None:
            (config_dir / name).write_bytes(data)

        return await loop.run_in_executor(
            None,
            _write_bytes,
            data,
        )

    async def destroy_kernel(
        self,
        kernel_id: KernelId,
        container_id: Optional[ContainerId],
    ) -> None:
        if container_id is None:
            return
        try:
            async with closing_async(Docker()) as docker:
                container = docker.containers.container(container_id)
                # The default timeout of the docker stop API is 10 seconds
                # to kill if container does not self-terminate.
                await container.stop()
        except DockerError as e:
            if e.status == HTTPStatus.CONFLICT and "is not running" in e.message:
                # already dead
                log.warning("destroy_kernel(k:{0}) already dead", kernel_id)
                await self.reconstruct_resource_usage()
            elif e.status == HTTPStatus.NOT_FOUND:
                # missing
                log.warning(
                    "destroy_kernel(k:{0}) kernel missing, forgetting this kernel", kernel_id
                )
                await self.reconstruct_resource_usage()
            else:
                log.exception("destroy_kernel(k:{0}) kill error", kernel_id)
                await self.error_monitor.capture_exception()

    async def clean_kernel(
        self,
        kernel_id: KernelId,
        container_id: Optional[ContainerId],
        restarting: bool,
    ) -> None:
        loop = current_loop()
        async with closing_async(Docker()) as docker:
            if container_id is not None:
                container = docker.containers.container(container_id)

                async def log_iter():
                    it = container.log(
                        stdout=True,
                        stderr=True,
                        follow=True,
                    )
                    async with aiotools.aclosing(it):
                        async for line in it:
                            yield line.encode("utf-8")

                try:
                    with timeout(60):
                        await self.collect_logs(kernel_id, container_id, log_iter())
                except DockerError as e:
                    if e.status == HTTPStatus.NOT_FOUND:
                        log.warning(
                            "container is already cleaned or missing (k:{}, cid:{})",
                            kernel_id,
                            container_id,
                        )
                    else:
                        raise
                except asyncio.TimeoutError:
                    log.warning(
                        "timeout for collecting container logs (k:{}, cid:{})",
                        kernel_id,
                        container_id,
                    )
                except Exception as e:
                    log.warning(
                        "error while collecting container logs (k:{}, cid:{})",
                        kernel_id,
                        container_id,
                        exc_info=e,
                    )

            kernel_obj = self.kernel_registry.get(kernel_id)
            if kernel_obj is not None:
                for domain_socket_proxy in kernel_obj.get("domain_socket_proxies", []):
                    if domain_socket_proxy.proxy_server.is_serving():
                        domain_socket_proxy.proxy_server.close()
                        await domain_socket_proxy.proxy_server.wait_closed()
                        try:
                            domain_socket_proxy.host_proxy_path.unlink()
                        except IOError:
                            pass

            if not self.local_config.debug.skip_container_deletion and container_id is not None:
                container = docker.containers.container(container_id)
                try:
                    with timeout(90):
                        await container.delete(force=True, v=True)
                except DockerError as e:
                    if e.status == HTTPStatus.CONFLICT and "already in progress" in e.message:
                        return
                    elif e.status == HTTPStatus.NOT_FOUND:
                        return
                    else:
                        log.exception(
                            "unexpected docker error while deleting container (k:{}, c:{})",
                            kernel_id,
                            container_id,
                        )
                except asyncio.TimeoutError:
                    log.warning("container deletion timeout (k:{}, c:{})", kernel_id, container_id)

            if not restarting:
                await _clean_scratch(
                    loop,
                    self.local_config.container.scratch_type,
                    self.local_config.container.scratch_root,
                    kernel_id,
                )
                if kernel_obj:
                    kernel = cast(DockerKernel, kernel_obj)
                    if kernel.network_driver != "bridge":
                        try:
                            plugin = self.network_plugin_ctx.plugins[kernel.network_driver]
                        except KeyError:
                            raise RuntimeError(
                                f"Network plugin {kernel.network_driver} not loaded!"
                            )
                        await plugin.leave_network(kernel)

    async def create_local_network(self, network_name: str) -> None:
        async with closing_async(Docker()) as docker:
            try:
                await docker.networks.get(network_name)
            except DockerError as e:
                if e.status == HTTPStatus.NOT_FOUND:
                    await docker.networks.create({
                        "Name": network_name,
                        "Driver": "bridge",
                        "Labels": {
                            "ai.backend.cluster-network": "1",
                        },
                    })
                else:
                    raise

    async def destroy_local_network(self, network_name: str) -> None:
        async with closing_async(Docker()) as docker:
            try:
                network = await docker.networks.get(network_name)
                await network.delete()
            except DockerError as e:
                if e.status == HTTPStatus.NOT_FOUND:
                    # skip silently if already removed/missing
                    pass
                else:
                    raise

    @preserve_termination_log
    async def monitor_docker_events(self):
        async def handle_action_start(
            session_id: SessionId, kernel_id: KernelId, evdata: Mapping[str, Any]
        ) -> None:
            await self.inject_container_lifecycle_event(
                kernel_id,
                session_id,
                LifecycleEvent.START,
                KernelLifecycleEventReason.NEW_CONTAINER_STARTED,
                container_id=ContainerId(evdata["Actor"]["ID"]),
            )

        async def handle_action_die(
            session_id: SessionId, kernel_id: KernelId, evdata: Mapping[str, Any]
        ) -> None:
            # When containers die, we immediately clean up them.
            reason = None
            kernel_obj = self.kernel_registry.get(kernel_id)
            if kernel_obj is not None:
                reason = kernel_obj.termination_reason
            try:
                exit_code = evdata["Actor"]["Attributes"]["exitCode"]
            except KeyError:
                exit_code = 255
            await self.inject_container_lifecycle_event(
                kernel_id,
                session_id,
                LifecycleEvent.CLEAN,
                reason or KernelLifecycleEventReason.SELF_TERMINATED,
                container_id=ContainerId(evdata["Actor"]["ID"]),
                exit_code=exit_code,
            )

        async def handle_action_oom(
            session_id: SessionId, kernel_id: KernelId, evdata: Mapping[str, Any]
        ) -> None:
            kernel_obj = self.kernel_registry.get(kernel_id, None)
            if kernel_obj is None:
                return
            await kernel_obj.notify_event(
                AgentEventData(
                    type="oom",
                    data={},
                )
            )

        while True:
            async with closing_async(Docker()) as docker:
                subscriber = docker.events.subscribe(create_task=True)
                try:
                    while True:
                        try:
                            # ref: https://docs.docker.com/engine/api/v1.40/#operation/SystemEvents
                            evdata = await subscriber.get()
                            if evdata is None:
                                # Break out to the outermost loop when the connection is closed
                                log.info(
                                    "monitor_docker_events(): "
                                    "restarting aiodocker event subscriber",
                                )
                                break
                            if evdata["Type"] != "container":
                                # Our interest is the container-related events
                                continue
                            container_name = evdata["Actor"]["Attributes"]["name"]
                            kernel_id = await get_kernel_id_from_container(container_name)
                            if kernel_id is None:
                                continue
                            if self.local_config.debug.log_docker_events and evdata["Action"] in (
                                "start",
                                "die",
                                "oom",
                            ):
                                log.debug(
                                    "docker-event: action={}, actor={}",
                                    evdata["Action"],
                                    evdata["Actor"],
                                )
                            session_id = SessionId(
                                UUID(evdata["Actor"]["Attributes"][LabelName.SESSION_ID])
                            )
                            match evdata["Action"]:
                                case "start":
                                    await asyncio.shield(
                                        self.docker_ptask_group.create_task(
                                            handle_action_start(session_id, kernel_id, evdata),
                                        )
                                    )
                                case "die":
                                    await asyncio.shield(
                                        self.docker_ptask_group.create_task(
                                            handle_action_die(session_id, kernel_id, evdata),
                                        )
                                    )
                                case "oom":
                                    await asyncio.shield(
                                        self.docker_ptask_group.create_task(
                                            handle_action_oom(session_id, kernel_id, evdata),
                                        )
                                    )
                        except asyncio.CancelledError:
                            # We are shutting down...
                            return
                        except Exception:
                            log.exception("monitor_docker_events(): unexpected error")
                finally:
                    await asyncio.shield(
                        self.docker_ptask_group.create_task(
                            docker.events.stop(),
                        )
                    )



import asyncio
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, override

from aiodocker.docker import Docker
from pydantic import ValidationError
from ruamel.yaml import YAML
from ruamel.yaml.error import YAMLError

from ai.backend.agent.exception import InvalidModelConfigurationError
from ai.backend.common.asyncio import closing_async
from ai.backend.common.config import ModelDefinition
from ai.backend.common.stage.types import (
    Provisioner,
    SpecGenerator,
)
from ai.backend.common.types import (
    MODEL_SERVICE_RUNTIME_PROFILES,
    RuntimeVariant,
    ServicePort,
    ServicePortProtocols,
    SessionTypes,
    VFolderMount,
)


@dataclass
class ModelServiceSpecCheckArgs:
    image_canonical: str
    model_vfolder_mount: VFolderMount
    session_type: SessionTypes
    runtime_variant: RuntimeVariant
    model_definition_path: Optional[str] = None

    def to_model_service_spec(self, image_command: Optional[str]) -> "ModelServiceSpec":
        return ModelServiceSpec(
            session_type=self.session_type,
            model_vfolder_mount=self.model_vfolder_mount,
            runtime_variant=self.runtime_variant,
            model_definition_path=self.model_definition_path,
            image_command=image_command,
        )


@dataclass
class ModelServiceSpec:
    session_type: SessionTypes
    model_vfolder_mount: VFolderMount
    runtime_variant: RuntimeVariant
    model_definition_path: Optional[str]
    image_command: Optional[str]


class ModelServiceSpecCheckProvisioner(Provisioner[ModelServiceSpecCheckArgs, ModelServiceSpec]):
    @property
    @override
    def name(self) -> str:
        return "docker-model-service-check"

    @override
    async def setup(self, spec: ModelServiceSpecCheckArgs) -> ModelServiceSpec:
        image_command = await self._extract_image_command(spec.image_canonical)
        if spec.runtime_variant != RuntimeVariant.CUSTOM and not image_command:
            raise InvalidModelConfigurationError(
                "image should have its own command when runtime variant is set to values other than CUSTOM!"
            )

        return spec.to_model_service_spec(image_command)

    async def _extract_image_command(self, image_canonical: str) -> Optional[str]:
        async with closing_async(Docker()) as docker:
            result = await docker.images.get(image_canonical)
            return result["Config"].get("Cmd")

    @override
    async def teardown(self, resource: ModelServiceSpec) -> None:
        # No teardown actions needed for this provisioner
        pass


class ModelServiceSpecGenerator(SpecGenerator[ModelServiceSpec]):
    def __init__(self, args: ModelServiceSpecCheckArgs) -> None:
        self._args = args
        self._provisioner = ModelServiceSpecCheckProvisioner()

    @override
    async def wait_for_spec(self) -> ModelServiceSpec:
        return await self._provisioner.setup(self._args)


@dataclass
class ModelServiceResult:
    model_definition: ModelDefinition
    service_ports: list[ServicePort]
    environ: dict[str, str]


class ModelServiceProvisioner(Provisioner[ModelServiceSpec, ModelServiceResult]):
    @property
    @override
    def name(self) -> str:
        return "docker-model-service"

    @override
    async def setup(self, spec: ModelServiceSpec) -> ModelServiceResult:
        model_definition = await self._get_model_definition(spec)
        if model_definition is None:
            raise InvalidModelConfigurationError(
                "Model definition is empty. Please check your model definition file"
            )

        environ: dict[str, str] = {}
        service_ports: list[ServicePort] = []
        for model in model_definition.models:
            environ["BACKEND_MODEL_NAME"] = model.name
            environ["BACKEND_MODEL_PATH"] = model.model_path
            if service := model.service:
                service_ports.append({
                    "name": f"{model.name}-{service.port}",
                    "protocol": ServicePortProtocols.PREOPEN,
                    "container_ports": (service.port,),
                    "host_ports": (None,),
                    "is_inference": True,
                })
        return ModelServiceResult(
            model_definition=model_definition,
            environ=environ,
            service_ports=service_ports,
        )

    async def _get_model_definition(self, spec: ModelServiceSpec) -> ModelDefinition:
        image_command = spec.image_command
        model_folder = spec.model_vfolder_mount
        match spec.runtime_variant:
            case RuntimeVariant.VLLM:
                return await self._get_model_definition_from_vllm(model_folder, image_command)
            case RuntimeVariant.HUGGINGFACE_TGI:
                return await self._get_model_definition_from_tgi(model_folder, image_command)
            case RuntimeVariant.NIM:
                return await self._get_model_definition_from_nim(model_folder, image_command)
            case RuntimeVariant.SGLANG:
                return await self._get_model_definition_from_sglang(model_folder, image_command)
            case RuntimeVariant.MODULAR_MAX:
                return await self._get_model_definition_from_modular_max(
                    model_folder, image_command
                )
            case RuntimeVariant.CMD:
                return await self._get_model_definition_from_cmd(model_folder, image_command)
            case RuntimeVariant.CUSTOM:
                return await self._get_model_definition_from_custom(
                    model_folder, spec.model_definition_path
                )

    async def _get_model_definition_from_vllm(
        self, model_folder: VFolderMount, image_command: Optional[str]
    ) -> ModelDefinition:
        _model = {
            "name": "vllm-model",
            "model_path": model_folder.kernel_path.as_posix(),
            "service": {
                "start_command": image_command,
                "port": MODEL_SERVICE_RUNTIME_PROFILES[RuntimeVariant.VLLM].port,
                "health_check": {
                    "path": MODEL_SERVICE_RUNTIME_PROFILES[
                        RuntimeVariant.VLLM
                    ].health_check_endpoint,
                },
            },
        }
        try:
            return ModelDefinition(models=[_model])  # type: ignore[arg-type,list-item]
        except ValidationError as e:
            raise InvalidModelConfigurationError(f"Invalid model definition for VLLM: {e}") from e

    async def _get_model_definition_from_tgi(
        self, model_folder: VFolderMount, image_command: Optional[str]
    ) -> ModelDefinition:
        _model = {
            "name": "tgi-model",
            "model_path": model_folder.kernel_path.as_posix(),
            "service": {
                "start_command": image_command,
                "port": MODEL_SERVICE_RUNTIME_PROFILES[RuntimeVariant.HUGGINGFACE_TGI].port,
                "health_check": {
                    "path": MODEL_SERVICE_RUNTIME_PROFILES[
                        RuntimeVariant.HUGGINGFACE_TGI
                    ].health_check_endpoint,
                },
            },
        }
        try:
            return ModelDefinition(models=[_model])  # type: ignore[arg-type,list-item]
        except ValidationError as e:
            raise InvalidModelConfigurationError(f"Invalid model definition for TGI: {e}") from e

    async def _get_model_definition_from_nim(
        self, model_folder: VFolderMount, image_command: Optional[str]
    ) -> ModelDefinition:
        _model = {
            "name": "nim-model",
            "model_path": model_folder.kernel_path.as_posix(),
            "service": {
                "start_command": image_command,
                "port": MODEL_SERVICE_RUNTIME_PROFILES[RuntimeVariant.NIM].port,
                "health_check": {
                    "path": MODEL_SERVICE_RUNTIME_PROFILES[
                        RuntimeVariant.NIM
                    ].health_check_endpoint,
                },
            },
        }
        try:
            return ModelDefinition(models=[_model])  # type: ignore[arg-type,list-item]
        except ValidationError as e:
            raise InvalidModelConfigurationError(f"Invalid model definition for NIM: {e}") from e

    async def _get_model_definition_from_sglang(
        self, model_folder: VFolderMount, image_command: Optional[str]
    ) -> ModelDefinition:
        _model = {
            "name": "sglang-model",
            "model_path": model_folder.kernel_path.as_posix(),
            "service": {
                "start_command": image_command,
                "port": MODEL_SERVICE_RUNTIME_PROFILES[RuntimeVariant.SGLANG].port,
                "health_check": {
                    "path": MODEL_SERVICE_RUNTIME_PROFILES[
                        RuntimeVariant.SGLANG
                    ].health_check_endpoint,
                },
            },
        }
        try:
            return ModelDefinition(models=[_model])  # type: ignore[arg-type,list-item]
        except ValidationError as e:
            raise InvalidModelConfigurationError(f"Invalid model definition for SGLang: {e}") from e

    async def _get_model_definition_from_modular_max(
        self, model_folder: VFolderMount, image_command: Optional[str]
    ) -> ModelDefinition:
        _model = {
            "name": "max-model",
            "model_path": model_folder.kernel_path.as_posix(),
            "service": {
                "start_command": image_command,
                "port": MODEL_SERVICE_RUNTIME_PROFILES[RuntimeVariant.MODULAR_MAX].port,
                "health_check": {
                    "path": MODEL_SERVICE_RUNTIME_PROFILES[
                        RuntimeVariant.MODULAR_MAX
                    ].health_check_endpoint,
                },
            },
        }
        try:
            return ModelDefinition(models=[_model])  # type: ignore[arg-type,list-item]
        except ValidationError as e:
            raise InvalidModelConfigurationError(
                f"Invalid model definition for Modular MAX: {e}"
            ) from e

    async def _get_model_definition_from_cmd(
        self, model_folder: VFolderMount, image_command: Optional[str]
    ) -> ModelDefinition:
        _model = {
            "name": "image-model",
            "model_path": model_folder.kernel_path.as_posix(),
            "service": {
                "start_command": image_command,
                "port": 8000,
            },
        }
        try:
            return ModelDefinition(models=[_model])  # type: ignore[arg-type,list-item]
        except ValidationError as e:
            raise InvalidModelConfigurationError(f"Invalid model definition for CMD: {e}") from e

    async def _get_model_definition_from_custom(
        self, model_folder: VFolderMount, model_definition_path: Optional[str]
    ) -> ModelDefinition:
        if model_definition_path:
            model_definition_candidates = [model_definition_path]
        else:
            model_definition_candidates = [
                "model-definition.yaml",
                "model-definition.yml",
            ]

        final_model_definition_path: Optional[Path] = None
        for filename in model_definition_candidates:
            if (Path(model_folder.host_path) / filename).is_file():
                final_model_definition_path = Path(model_folder.host_path) / filename
                break

        if not final_model_definition_path:
            raise InvalidModelConfigurationError(
                f"Model definition file ({' or '.join(model_definition_candidates)}) does not exist under VFolder"
                f" {model_folder.name} (ID {model_folder.vfid})",
            )
        try:
            model_definition_yaml = await asyncio.get_running_loop().run_in_executor(
                None, final_model_definition_path.read_text
            )
        except FileNotFoundError as e:
            raise InvalidModelConfigurationError(
                "Model definition file (model-definition.yml) does not exist under"
                f" vFolder {model_folder.name} (ID {model_folder.vfid})",
            ) from e
        try:
            yaml = YAML()
            raw_definition = yaml.load(model_definition_yaml)
        except YAMLError as e:
            raise InvalidModelConfigurationError(f"Invalid YAML syntax: {e}") from e
        try:
            return ModelDefinition(**raw_definition)
        except ValidationError as e:
            raise InvalidModelConfigurationError(f"Invalid model definition: {e}") from e

    @override
    async def teardown(self, resource: ModelServiceResult) -> None:
        return



from __future__ import annotations

import enum
from collections import defaultdict
from collections.abc import Iterable
from typing import TYPE_CHECKING, Optional, Sequence

import attr
import networkx as nx

from ai.backend.common.types import DeviceName

if TYPE_CHECKING:
    from .resources import AbstractComputeDevice


class AffinityPolicy(enum.Enum):
    PREFER_SINGLE_NODE = 0
    INTERLEAVED = 1


@attr.define()
class AffinityHint:
    """
    Represents the affinity hint for realization of an entire resource slot for a single kernel.

    The devices field starts with None and gets replaced with the devices allocated in the prior resource slot
    as the allocation proceeds with the next resource slot.
    """

    devices: Optional[Sequence[AbstractComputeDevice]]
    affinity_map: AffinityMap
    policy: AffinityPolicy


class AffinityMap(nx.Graph):
    """
    Represents the NUMA distance matrix of all device pairs from all compute device plugins.
    """

    def __init__(self) -> None:
        self.min_weight = 0
        self.max_weight = 0
        super().__init__()

    def get_largest_device_cluster_with_lowest_distance_from_src_device(
        self,
        device_name: DeviceName,
        src_device: AbstractComputeDevice,
    ) -> Sequence[Sequence[AbstractComputeDevice]]:
        distance_sets: dict[int, nx.Graph] = defaultdict(nx.Graph)
        for v in self.neighbors(src_device):
            if v.device_name == device_name:
                weight = self.edges[src_device, v]["weight"]
                distance_sets[weight].add_edge(src_device, v, weight=weight)
        device_cluster_list = []
        for distance, device_set in distance_sets.items():
            components = nx.connected_components(device_set)
            for component in components:
                device_cluster_list.append((distance, component))
        # sort by: low distance first, large component first
        device_cluster_list.sort(key=lambda item: (item[0], -len(item[1])))
        largest_components: list[list[AbstractComputeDevice]] = []
        for distance, subgraph in device_cluster_list:
            largest_component: list[AbstractComputeDevice] = []
            for device in subgraph:
                if device == src_device:
                    continue
                largest_component.append(device)
            largest_components.append(largest_component)
        return largest_components

    def get_device_clusters_with_lowest_distance(
        self,
        device_name: DeviceName,
    ) -> Sequence[Sequence[AbstractComputeDevice]]:
        """
        Get the device clusters having the minimum distance from each other filtered by device_name.

        This method is intended to be called for the first resource slot of a single allocation run.
        In most cases, it will be accelerators as defined in the device allocation order
        (see :attr:`ai.backend.agent.config.unified.ResourceConfig.allocation_order`)
        """
        device_cluster_list: list[tuple[int, set[AbstractComputeDevice]]] = []
        for weight in [self.min_weight]:
            subgraph = nx.subgraph_view(
                self,
                filter_node=lambda u: u.device_name == device_name,
                filter_edge=lambda u, v: self.edges[u, v]["weight"] == weight,
            )
            components: Iterable[set[AbstractComputeDevice]] = nx.connected_components(subgraph)
            for component in components:
                device_cluster_list.append((weight, component))
        # sort by: low distance first, large component first
        device_cluster_list.sort(key=lambda item: (item[0], -len(item[1])))
        return [[*device_set] for distance, device_set in device_cluster_list]

    def get_distance_ordered_neighbors(
        self,
        src_devices: Sequence[AbstractComputeDevice],
        device_name: DeviceName,
    ) -> tuple[Sequence[Sequence[AbstractComputeDevice]], Sequence[AbstractComputeDevice]]:
        """
        Get the list of neighbor device clusters and their distance from the given src_devices
        with the same name.

        This method is intended to be called for non-first resource slots of a single allocation run.
        It is used to prioritize the closest devices of the current resource slot
        from the already allocated devices (src_devices) of the previous resource slot.

        Example:
            Given a 4-core dual socket system with two GPUs per socket:

            If the prior allocator has assigned (gpu0@node0, gpu2@node1),
            it will return:
               primary: cpu0-3@node0, cpu4-7@node1 [interleaved]
               secondary: (empty) [fill-remaining]

            If the prior allocator has assigned (gpu0@node0, gpu1@node0),
            it will return:
              primary: cpu0-3@node0 [interleaved]
              secondary: cpu4-7@node1 [fill-remaining]
        """
        if not src_devices:
            raise RuntimeError(
                "This method must be called when there are previously allocated devices in a single allocation run."
            )
        if next(iter(src_devices)).device_name == device_name:
            raise RuntimeError(
                "This is a logic error trying to allocate the same resource slots twice in a single allocation run."
            )
        src_numa_nodes = set(src_device.numa_node for src_device in src_devices)
        primary_sets = []
        secondary_set = set()
        for numa_node in src_numa_nodes:
            primary_set = set()
            for u in self.nodes:
                if u.device_name == device_name:
                    if u.numa_node == numa_node:
                        primary_set.add(u)
                    else:
                        secondary_set.add(u)
            primary_sets.append(primary_set)
        for primary_set in primary_sets:
            secondary_set -= primary_set
        return [*(list(primary_set) for primary_set in primary_sets)], list(secondary_set)

    @classmethod
    def build(cls, devices: Sequence[AbstractComputeDevice]) -> AffinityMap:
        # TODO: allow compute plugins to customize distance calculation
        g = cls()
        max_weight = 0
        min_weight = 0
        devices_copy = list(devices)
        while devices_copy:
            device1 = devices_copy.pop(0)
            g.add_edge(device1, device1, weight=0)
            for device2 in devices_copy:
                # weight = abs((max(0, device2.numa_node or 0)) - max(0, (device1.numa_node or 0)))
                weight = 0 if device1.numa_node == device2.numa_node else 1
                if max_weight < weight:
                    max_weight = weight
                if min_weight > weight:
                    min_weight = weight
                g.add_edge(device1, device2, weight=weight)
        g.max_weight = max_weight
        g.min_weight = min_weight
        return g



from __future__ import annotations

import asyncio
import hashlib
import io
import ipaddress
import logging
import platform
import re
import resource
from decimal import Decimal
from pathlib import Path
from typing import (
    Any,
    AsyncContextManager,
    List,
    Mapping,
    MutableMapping,
    NamedTuple,
    Optional,
    Protocol,
    Tuple,
    Type,
    TypedDict,
    TypeVar,
    Union,
    overload,
)
from uuid import UUID

import aiodocker
import trafaret as t
from aiodocker.docker import DockerContainer
from typing_extensions import Final

from ai.backend.common import identity
from ai.backend.common.cgroup import (
    get_cgroup_of_pid,
    get_container_id_of_cgroup,
    get_container_pids,
)
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.json import dump_json_str
from ai.backend.common.types import PID, ContainerId, ContainerPID, HostPID, KernelId
from ai.backend.common.utils import current_loop
from ai.backend.logging import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

IPNetwork = Union[ipaddress.IPv4Network, ipaddress.IPv6Network]
IPAddress = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]

InOtherContainerPID: Final = ContainerPID(PID(-2))
NotContainerPID: Final = ContainerPID(PID(-1))
NotHostPID: Final = HostPID(PID(-1))


class SupportsAsyncClose(Protocol):
    async def close(self) -> None: ...


_SupportsAsyncCloseT = TypeVar("_SupportsAsyncCloseT", bound=SupportsAsyncClose)


class closing_async(AsyncContextManager[_SupportsAsyncCloseT]):
    """
    contextlib.closing calls close(), and aiotools.aclosing() calls aclose().
    This context manager calls close() as a coroutine.
    """

    def __init__(self, obj: _SupportsAsyncCloseT) -> None:
        self.obj = obj

    async def __aenter__(self) -> _SupportsAsyncCloseT:
        return self.obj

    async def __aexit__(self, *exc_info) -> None:
        await self.obj.close()


def generate_local_instance_id(hint: str) -> str:
    return hashlib.md5(hint.encode("utf-8")).hexdigest()[:12]


def get_arch_name() -> str:
    ret = platform.machine().lower()
    aliases = {
        "arm64": "aarch64",  # macOS with LLVM
        "amd64": "x86_64",  # Windows/Linux
        "x64": "x86_64",  # Windows
        "x32": "x86",  # Windows
        "i686": "x86",  # Windows
    }
    return aliases.get(ret, ret)


def update_nested_dict(dest: MutableMapping, additions: Mapping) -> None:
    for k, v in additions.items():
        if k not in dest:
            dest[k] = v
        else:
            if isinstance(dest[k], MutableMapping):
                assert isinstance(v, MutableMapping)
                update_nested_dict(dest[k], v)
            elif isinstance(dest[k], List):
                assert isinstance(v, List)
                dest[k].extend(v)
            else:
                dest[k] = v


def numeric_list(s: str) -> List[int]:
    return [int(p) for p in s.split()]


def remove_exponent(num: Decimal) -> Decimal:
    return num.quantize(Decimal(1)) if num == num.to_integral() else num.normalize()


@overload
def read_sysfs(
    path: Union[str, Path], type_: Type[bool], default: Optional[bool] = None
) -> bool: ...


@overload
def read_sysfs(path: Union[str, Path], type_: Type[int], default: Optional[int] = None) -> int: ...


@overload
def read_sysfs(
    path: Union[str, Path], type_: Type[float], default: Optional[float] = None
) -> float: ...


@overload
def read_sysfs(path: Union[str, Path], type_: Type[str], default: Optional[str] = None) -> str: ...


def read_sysfs(path: Union[str, Path], type_: Type[Any], default: Optional[Any] = None) -> Any:
    def_vals: Mapping[Any, Any] = {
        bool: False,
        int: 0,
        float: 0.0,
        str: "",
    }
    if type_ not in def_vals:
        raise TypeError("unsupported conversion type from sysfs content")
    if default is None:
        default = def_vals[type_]
    try:
        raw_str = Path(path).read_text().strip()
        if type_ is bool:
            return t.ToBool().check(raw_str)
        else:
            return type_(raw_str)
    except IOError:
        return default


async def read_tail(path: Path, nbytes: int) -> bytes:
    file_size = path.stat().st_size

    def _read_tail() -> bytes:
        with open(path, "rb") as f:
            f.seek(max(file_size - nbytes, 0), io.SEEK_SET)
            return f.read(nbytes)

    loop = current_loop()
    return await loop.run_in_executor(None, _read_tail)


async def get_kernel_id_from_container(val: Union[str, DockerContainer]) -> Optional[KernelId]:
    if isinstance(val, DockerContainer):
        if "Name" not in val._container:
            await val.show()
        name = val["Name"]
    elif isinstance(val, str):
        name = val
    name = name.lstrip("/")
    if not name.startswith("kernel."):
        return None
    try:
        return KernelId(UUID(name.rsplit(".", 2)[-1]))
    except (IndexError, ValueError):
        return None


async def get_subnet_ip(etcd: AsyncEtcd, network: str, fallback_addr: str = "0.0.0.0") -> str:
    raw_subnet = await etcd.get(f"config/network/subnet/{network}")
    if raw_subnet is None:
        addr = fallback_addr
    else:
        subnet = ipaddress.ip_network(raw_subnet)
        if subnet.prefixlen == 0:
            addr = fallback_addr
        else:
            local_ipaddrs = [*identity.fetch_local_ipaddrs(subnet)]
            log.debug("get_subnet_ip(): subnet {} candidates: {}", subnet, local_ipaddrs)
            if local_ipaddrs:
                addr = str(local_ipaddrs[0])
            else:
                addr = fallback_addr
    return addr


class ProcessInfo(NamedTuple):
    pid: int
    command: str


async def get_host_process_table(
    docker: aiodocker.Docker,
    container_id: str,
) -> List[ProcessInfo]:
    result = await docker._query_json(f"containers/{container_id}/top", method="GET")
    procs = result["Processes"]
    return [ProcessInfo(int(x[1]), x[7]) for x in procs]


async def get_container_process_table(
    docker: aiodocker.Docker,
    container_id: str,
) -> List[ProcessInfo]:
    # Get process table from inside container (execute 'ps -aux' command from container).
    # Filter processes which have exactly the same COMMAND like above.
    result = await docker._query_json(
        f"containers/{container_id}/exec",
        method="POST",
        data={
            "AttachStdin": False,
            "AttachStdout": True,
            "AttachStderr": True,
            "Cmd": ["ps", "aux"],
        },
    )
    exec_id = result["Id"]
    async with docker._query(
        f"exec/{exec_id}/start",
        method="POST",
        headers={"content-type": "application/json"},
        data=dump_json_str({
            "Stream": False,  # get response immediately
            "Detach": False,
            "Tty": False,
        }),
    ) as resp:
        result = await resp.read()
        result = result.decode("latin-1").strip().split("\n")
    result = list(map(lambda x: x.split(), result))
    head = result[0]
    procs = result[1:]
    pid_idx, cmd_idx = head.index("PID"), head.index("COMMAND")
    return [ProcessInfo(int(r[pid_idx]), " ".join(r[cmd_idx:])) for r in procs]


async def host_pid_to_container_pid(container_id: str, host_pid: HostPID) -> ContainerPID:
    kernel_ver = Path("/proc/version").read_text()
    if m := re.match(r"Linux version (\d+)\.(\d+)\..*", kernel_ver):  # noqa
        kernel_ver_tuple: Tuple[str, str] = m.groups()  # type: ignore
        if kernel_ver_tuple < ("4", "1"):
            # TODO: this should be deprecated when the minimun supported Linux kernel will be 4.1.
            #
            # In CentOs 7, NSPid is not accesible since it is supported from Linux kernel >=4.1.
            # We provide alternative, although messy, way for older Linux kernels. Below describes
            # the logic briefly:
            #   * Obtain information on all the processes inside the target container,
            #     which contains host PID, by docker top API (containers/<container-id>/top).
            #     - Get the COMMAND of the target process (by using host_pid).
            #     - Filter host processes which have the exact same COMMAND.
            #   * Obtain information on all the processes inside the target container,
            #     which contains container PID, by executing "ps -aux" command from inside the container.
            #     - Filter container processes which have the exact same COMMAND.
            #   * Get the index of the target process from the host process table.
            #   * Use the index to get the target process from the container process table, and get PID.
            #     - Since docker top and ps -aux both displays processes in the order of PID, we
            #       can safely assume that the order of the processes from both tables are the same.
            #
            # Example host and container process table:
            #
            # [
            #   ['devops', '15454', '12942', '99', '15:36', 'pts/1', '00:00:08', 'python mnist.py'],
            #   ... (processes with the same COMMAND)
            # ]
            #
            # [
            #   ['work', '227', '121', '4.6', '22408680', '1525428', 'pts/1', 'Rl+', '06:36', '0:08',
            #    'python', 'mnist.py'],
            #   ... (processes with the same COMMAND)
            # ]
            try:
                docker = aiodocker.Docker()
                # Get process table from host (docker top information). Filter processes which have
                # exactly the same COMMAND as with target host process.
                host_procs = await get_host_process_table(docker, container_id)
                cmd = [x.command for x in host_procs if x.pid == int(host_pid)][0]
                host_pids = [x.pid for x in host_procs if x.command == cmd]

                # When there are multiple processes which have the same COMMAND, just get the index of
                # the target host process and apply it with the container table. Since ps and docker top
                # both displays processes ordered by PID, we can expect those two tables have same
                # order of processes.
                process_idx = host_pids.index(int(host_pid))

                container_procs = await get_container_process_table(docker, container_id)
                container_pids = [x.pid for x in container_procs if x.command == cmd]

                container_pid = ContainerPID(PID(container_pids[process_idx]))
                log.debug("host pid {} is mapped to container pid {}", host_pid, container_pid)
                return container_pid
            except asyncio.CancelledError:
                raise
            except (IndexError, KeyError, aiodocker.exceptions.DockerError):
                return NotContainerPID
            finally:
                await docker.close()

    try:
        cgroup = get_cgroup_of_pid("pids", host_pid)
        cgroup_container_id = get_container_id_of_cgroup(cgroup)
        if cgroup_container_id is None:
            return NotContainerPID
        if cgroup_container_id == container_id:
            for line in Path(f"/proc/{host_pid}/status").read_text().splitlines():
                key, value = line.split(":\t", 1)
                if key == "NSpid":
                    pid = value.split()[1]
                    return ContainerPID(PID(int(pid)))
        return InOtherContainerPID
    except OSError:
        return NotContainerPID


async def container_pid_to_host_pid(container_id: str, container_pid: ContainerPID) -> HostPID:
    kernel_ver = Path("/proc/version").read_text()
    if m := re.match(r"Linux version (\d+)\.(\d+)\..*", kernel_ver):  # noqa
        kernel_ver_tuple: Tuple[str, str] = m.groups()  # type: ignore
        if kernel_ver_tuple < ("4", "1"):
            # reverse implementation of host_pid_to_container_pid().
            try:
                docker = aiodocker.Docker()
                container_procs = await get_container_process_table(docker, container_id)
                cmd = [x.command for x in container_procs if x.pid == int(container_pid)][0]
                container_pids = [x.pid for x in container_procs if x.command == cmd]
                process_idx = container_pids.index(int(container_pid))

                host_procs = await get_host_process_table(docker, container_id)
                host_pids = [x.pid for x in host_procs if x.command == cmd]

                host_pid = HostPID(PID(host_pids[process_idx]))
                log.debug("container pid {} is mapped to host pid {}", container_pid, host_pid)
                return host_pid
            except asyncio.CancelledError:
                raise
            except (IndexError, KeyError, aiodocker.exceptions.DockerError):
                return NotHostPID
            finally:
                await docker.close()

    try:
        cgtasks = await get_container_pids(ContainerId(container_id))
        for pid in cgtasks:
            proc_path = Path(f"/proc/{pid}/status")
            proc_status = {
                k: v
                for k, v in map(lambda line: line.split(":\t"), proc_path.read_text().splitlines())
            }
            nspids = proc_status["NSpid"].split()
            if nspids[1] == str(container_pid):
                return HostPID(PID(pid))
        return NotHostPID
    except (ValueError, KeyError, IOError):
        return NotHostPID


class DockerUlimitEntry(TypedDict):
    Name: str
    Soft: int
    Hard: int


def get_safe_ulimit(name: str, desired_soft: int, desired_hard: int) -> DockerUlimitEntry:
    """
    Get ulimit values that don't exceed the host's system limits to avoid
    'operation not permitted' errors from Docker.
    """
    try:
        # Map ulimit names to resource constants
        ulimit_map = {
            "nofile": resource.RLIMIT_NOFILE,
            "memlock": resource.RLIMIT_MEMLOCK,
            "nproc": resource.RLIMIT_NPROC,
            "fsize": resource.RLIMIT_FSIZE,
            "core": resource.RLIMIT_CORE,
            "stack": resource.RLIMIT_STACK,
            "data": resource.RLIMIT_DATA,
            "rss": resource.RLIMIT_RSS,
        }

        if name not in ulimit_map:
            # If we don't know how to check this limit, use desired values
            return {"Name": name, "Soft": desired_soft, "Hard": desired_hard}

        current_soft_limit, current_hard_limit = resource.getrlimit(ulimit_map[name])
        infinity_val = resource.RLIM_INFINITY

        if desired_soft == -1:
            safe_soft = current_soft_limit if current_soft_limit != infinity_val else -1
        else:
            if current_soft_limit == infinity_val:
                safe_soft = desired_soft
            else:
                safe_soft = min(desired_soft, current_soft_limit)

        if desired_hard == -1:
            safe_hard = current_hard_limit if current_hard_limit != infinity_val else -1
        else:
            if current_hard_limit == infinity_val:
                safe_hard = desired_hard
            else:
                safe_hard = min(desired_hard, current_hard_limit)

        # Log if we had to adjust the limits
        if safe_soft != desired_soft or safe_hard != desired_hard:
            log.debug(
                f"Adjusted ulimit {name}: desired=({desired_soft}, {desired_hard}) "
                f"-> safe=({safe_soft}, {safe_hard}) (system limits: {current_soft_limit}, {current_hard_limit})"
            )

        return {"Name": name, "Soft": safe_soft, "Hard": safe_hard}

    except (OSError, ValueError, AttributeError) as e:
        # If we can't get system limits, log and use desired values
        log.warning(f"Could not get system ulimit for {name}: {e}")
        return {"Name": name, "Soft": desired_soft, "Hard": desired_hard}




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/ai/backend/agent:dist

from setuptools import setup

setup(**{
    'author': 'Lablup Inc. and contributors',
    'classifiers': [
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.13',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
    ],
    'description': 'Backend.AI Agent',
    'entry_points': {
        'backendai_cli_v10': [
            'ag = ai.backend.agent.cli.__main__:main',
            'ag.start-server = ai.backend.agent.server:main',
        ],
        'backendai_network_agent_v1': [
            'overlay = ai.backend.agent.docker.intrinsic:OverlayNetworkPlugin',
            'host = ai.backend.agent.docker.intrinsic:HostNetworkPlugin',
        ],
    },
    'install_requires': (
        'aiodocker==0.24.0',
        'aiofiles~=24.1.0',
        'aiohttp_cors~=0.8.1',
        'aiohttp~=3.12.15',
        'aiomonitor~=0.7.0',
        'aiotools~=2.2.3',
        'async_timeout~=4.0',
        'attrs>=25.3',
        """backend.ai-cli==25.16.0
""",
        """backend.ai-common==25.16.0
""",
        """backend.ai-kernel-binary==25.16.0
""",
        """backend.ai-kernel-helper==25.16.0
""",
        """backend.ai-kernel==25.16.0
""",
        'backend.ai-krunner-static-gnu==4.4.0',
        """backend.ai-logging==25.16.0
""",
        """backend.ai-plugin==25.16.0
""",
        'cachetools~=5.5.0',
        'callosum~=1.0.3',
        'cattrs~=24.1.1',
        'click~=8.1.7',
        'etcd-client-py~=0.4.1',
        'janus~=2.0',
        'kubernetes-asyncio~=33.3.0',
        'kubernetes~=33.1.0',
        'more-itertools~=10.5.0',
        'networkx~=3.3.0',
        'prometheus-client~=0.21.1',
        'psutil~=7.0',
        'pydantic[email]~=2.11.3',
        'pyzmq~=26.4',
        'ruamel.yaml~=0.18.10',
        'setproctitle~=1.3.5',
        'setuptools~=80.0.0',
        'tenacity>=9.0',
        'tomlkit~=0.13.2',
        'trafaret~=2.1',
        'types-aiofiles',
        'types-cachetools',
        'types-psutil',
        'types-setuptools',
        'typing_extensions~=4.11',
        'uvloop~=0.22.1; sys_platform != "Windows"',
    ),
    'license': 'LGPLv3',
    'long_description': """# Backend.AI Agent

## Purpose

The Agent is a compute node component responsible for container lifecycle management. It runs computing sessions as containers, monitors resource usage, and reports status to the Manager.

## Key Responsibilities

### 1. Container Lifecycle Management
- Create and start containers from session specifications
- Initialize kernels and execute runtime configuration
- Monitor container health and status
- Delete containers upon session termination
- Handle container restarts and error recovery

### 2. Resource Monitoring
- Track CPU, memory, GPU usage per container
- Monitor disk I/O and network traffic
- Collect accelerator metrics (CUDA, ROCm, TPU)
- Report resource availability to Manager
- Enforce resource limits via cgroups

### 3. Local Storage Management
- Manage scratch storage for temporary files
- Mount virtual folders (vfolders) to containers
- Handle file permissions and ownership
- Clean up storage after session termination

### 4. Service Port Management
- Expose container service ports (Jupyter, SSH, TensorBoard, etc.)
- Forward traffic from App Proxy to containers
- Manage port allocation from configured ranges
- Handle SSL/TLS termination when needed

### 5. Agent Status Reporting
- Periodically report agent status to Manager
- Send heartbeat signals indicating availability
- Update occupied and available resource slots
- Report errors and exceptional conditions

## Architecture

```
┌────────────────────────────────────────┐
│         Agent Server (agent.py)        │
├────────────────────────────────────────┤
│    Container Backend (docker/)         │  ← Docker
├────────────────────────────────────────┤
│  Resource Monitor (watcher/ stats.py)  │  ← CPU, GPU, Memory
├────────────────────────────────────────┤
│   Storage Manager (scratch.py fs.py)   │  ← Local and mounted storage
├────────────────────────────────────────┤
│      Plugin System (plugin/)           │  ← Accelerator plugins
└────────────────────────────────────────┘
```

## Directory Structure

```
agent/
├── docker/              # Docker container backend
│   ├── agent.py        # Docker-specific agent implementation
│   └── resources.py    # Docker resource management
├── watcher/             # Resource monitoring
│   ├── cpu.py          # CPU monitoring
│   ├── mem.py          # Memory monitoring
│   └── gpu.py          # GPU monitoring
├── plugin/              # Accelerator plugins
│   ├── cuda.py         # NVIDIA CUDA support
│   ├── rocm.py         # AMD ROCm support
│   └── tpu.py          # Google TPU support
├── observer/            # Metrics observers
│   └── stat.py         # Statistics collection
├── cli/                 # CLI commands
├── config/              # Configuration
├── agent.py             # Main agent logic
├── server.py            # RPC server entry point
├── kernel.py            # Kernel (container) management
├── resources.py         # Resource allocation
├── stats.py             # Statistics aggregation
├── scratch.py           # Scratch storage management
└── fs.py                # Filesystem utilities
```

## Core Concepts

### Kernels
Kernels represent running containers executing computing sessions:
- **Kernel ID**: Unique identifier for the container
- **Session ID**: Associated session in Manager
- **Image**: Container image (e.g., `lablup/python:3.11-ubuntu20.04`)
- **Resources**: Allocated CPU, memory, GPU slots
- **Status**: PREPARING, RUNNING, RESTARTING, TERMINATING, TERMINATED
- **Service Ports**: Exposed ports for services (Jupyter, SSH, etc.)

### Container Backend
The Agent supports multiple container runtimes:
- **Docker**: Standard Docker containers

The Docker backend implements these interfaces:
- `create_kernel()`: Create new container
- `destroy_kernel()`: Remove container
- `restart_kernel()`: Restart container
- `get_kernel_status()`: Query container status
- `execute_code()`: Execute code in container

### Resource Allocation
Resources are allocated from the agent's capacity:
- **CPU Slots**: Measured in cores (e.g., 2.0 cores)
- **Memory Slots**: Measured in bytes (e.g., 4GB)
- **GPU Slots**: Number of GPU devices (e.g., 1 GPU)
- **Accelerator Slots**: Custom accelerator resources

The Agent tracks:
- **Total Capacity**: Maximum available resources
- **Occupied Slots**: Currently allocated resources
- **Available Slots**: Remaining resources for new sessions

### Scratch Storage
Each kernel receives local scratch storage:
- **Location**: `/tmp/backend.ai/scratches/{kernel_id}`
- **Quota**: Configurable size limit per kernel
- **Cleanup**: Automatically removed after kernel termination
- **Performance**: Local SSD for fast I/O

### Virtual Folder Mounting
VFolders are mounted to containers at runtime:
- **Mount Point**: `/home/work/{vfolder_name}`
- **Permissions**: RO, RW, or RW-DELETE
- **Backend**: Storage Proxy manages actual storage
- **Protocol**: NFS, SMB, or direct mount

## Resource Monitoring

### CPU Monitoring
- Track per-container CPU usage via cgroups
- Measure CPU time in user and system modes
- Calculate CPU utilization percentages
- Enforce CPU quotas and limits

### Memory Monitoring
- Track RSS (Resident Set Size) per container
- Measure cache and swap usage
- Detect OOM (Out-of-Memory) conditions
- Enforce memory limits via cgroups

### GPU Monitoring
- Query NVIDIA GPUs via NVML (nvidia-ml-py)
- Query AMD GPUs via ROCm SMI
- Track GPU utilization and memory usage
- Measure GPU temperature and power consumption

### Disk I/O Monitoring
- Track read/write operations per container
- Measure I/O bandwidth usage
- Monitor disk space consumption
- Enforce I/O throttling when configured

## Plugin System

The Agent uses plugins for accelerator support:

### CUDA Plugin
- Detect NVIDIA GPUs via `nvidia-smi`
- Allocate GPU devices to containers
- Set `CUDA_VISIBLE_DEVICES` environment variable
- Monitor GPU metrics via NVML

### ROCm Plugin
- Detect AMD GPUs via `rocm-smi`
- Allocate GPU devices to containers
- Set `HIP_VISIBLE_DEVICES` environment variable
- Monitor GPU metrics via ROCm

### TPU Plugin
- Detect Google TPUs
- Configure TPU access for TensorFlow
- Monitor TPU utilization

## Communication Protocols

### Manager → Agent (ZeroMQ RPC)
- **Port**: 6011 (default)
- **Protocol**: ZeroMQ request-response
- **Operations**:
  - `create_kernel`: Create new container
  - `destroy_kernel`: Terminate container
  - `restart_kernel`: Restart container
  - `execute_code`: Execute code in container
  - `get_status`: Query agent and kernel status

### Agent → Manager (HTTP Watcher API)
- **Port**: 6009 (default)
- **Protocol**: HTTP
- **Operations**:
  - Heartbeat signals
  - Resource usage reporting
  - Kernel status updates
  - Error notifications

### Agent → Storage Proxy
- **Protocol**: HTTP or NFS/SMB
- **Operations**:
  - Mount vfolder
  - Unmount vfolder
  - Query vfolder metadata

## Container Execution Flow

```
1. Manager sends create_kernel RPC
   ↓
2. Agent validates resource availability
   ↓
3. Agent pulls container image (if needed)
   ↓
4. Agent creates scratch directory
   ↓
5. Agent mounts vfolders via Storage Proxy
   ↓
6. Agent creates container with resources
   ↓
7. Agent starts container and runs init script
   ↓
8. Agent registers service ports
   ↓
9. Agent reports kernel status to Manager
   ↓
10. Container runs until termination
   ↓
11. Agent cleans up resources upon termination
```

## Service Ports

Containers can expose service ports:
- **Jupyter Notebook**: Port 8080 (HTTP)
- **Jupyter Lab**: Port 8090 (HTTP)
- **SSH**: Port 2200 (TCP)
- **TensorBoard**: Port 6006 (HTTP)
- **Custom Services**: User-defined ports

Service ports are:
- **Allocated** from configured range (default 30000-31000)
- **Registered** with App Proxy for external access
- **Forwarded** from agent host to container
- **Cleaned up** upon container termination

## Configuration

See `configs/agent/halfstack.toml` for configuration file examples.

### Key Configuration Items

**Agent Basic Settings**:
- Agent ID and region information
- Resource slot definitions
- Container runtime configuration (Docker)

**Storage Settings**:
- Scratch storage root path
- Per-session quota

**Service Port Settings**:
- Container port range

**Resource Monitoring**:
- Watcher interval settings

## Infrastructure Dependencies

### Required Infrastructure

#### Container Runtime
- **Purpose**: Container creation and management
- **Supported Runtimes**: Docker

#### Redis (Event and State Management)
- **Purpose**:
  - Send heartbeat events (for Manager registration)
  - Update agent status
  - Manage background tasks
- **Note**: Agent registers with Manager via Redis and receives RPC commands from Manager.

#### Manager Connection
- **Protocol**: ZeroMQ RPC (6011), HTTP Watcher API (6009)
- **Purpose**: Receive session commands from Manager, report status
- **Note**: Agent only receives commands from Manager, not directly from users or other components.

#### etcd (Global Configuration)
- **Purpose**:
  - Retrieve global configuration (container registry, storage volumes, etc.)
  - Auto-discover Manager address

### Optional Infrastructure (Observability)

Optional infrastructure for Agent monitoring.

#### Prometheus (Metrics Collection)
- **Purpose**:
  - Monitor agent resource utilization
  - Collect kernel (container) metrics
  - Track heartbeat and health status
- **HTTP Service Port**: 6003 (separate from RPC port 6011)
- **Exposed Endpoint**: `http://localhost:6003/metrics`
- **Key Metrics**:
  - `backendai_agent_heartbeat` - Last heartbeat timestamp
  - `backendai_agent_cpu_usage` - CPU usage percentage
  - `backendai_agent_mem_usage` - Memory usage (bytes)
  - `backendai_agent_gpu_usage` - GPU usage percentage
  - `backendai_kernel_count` - Number of running kernels
- **Service Discovery**: Automatically registered via Manager's HTTP SD
  - Agent auto-registers with Manager at startup
  - Manager provides agent information to Prometheus SD

#### Loki (Log Aggregation)
- **Purpose**:
  - Centralized agent log collection
  - Aggregate kernel execution logs
  - Track errors and debugging information
- **Log Transmission Methods**:
  - Direct HTTP API usage
  - Use Promtail or Fluent Bit
- **Log Labels**:
  - `agent_id` - Agent identifier
  - `kernel_id` - Kernel identifier
  - `level` - Log level

#### Container Runtime Metrics
- **Docker**: Container metrics via cAdvisor or Docker API
- **cgroups**: Direct cgroups filesystem reading

### Container Runtime Requirements

#### Docker
- **Minimum Version**: Docker 20.10+
- **Required Features**:
  - cgroups v2 support (resource limits)
  - GPU support (NVIDIA Docker Runtime or CDI)

### Storage Requirements

#### Local Scratch Storage
- **Purpose**: Temporary files, build cache
- **Recommended Specifications**:
  - Fast SSD (NVMe recommended)
  - Minimum 100GB free space
  - Configurable per-session quota

#### VFolder Mount
- **Protocol**: NFS, SMB, direct mount
- **Dependencies**: Storage Proxy or direct storage access
- **Recommendations**:
  - High-performance network (10GbE or higher)
  - Low-latency storage

### Halfstack Configuration

**Recommended**: Use the `./scripts/install-dev.sh` script for development environment setup.

#### Starting Development Environment
```bash
# Setup development environment via script (recommended)
./scripts/install-dev.sh

# Start Agent
./backend.ai ag start-server
```

#### Observability Integration
Agent automatically registers with Prometheus via Manager's Service Discovery.

## Metrics and Monitoring

### Prometheus Metrics
Agent exposes metrics:

#### RPC Related Metrics
Metrics related to processing RPC commands received from Manager.

**`backendai_rpc_requests`** (Counter)
- **Description**: Total RPC requests processed by the agent
- **Labels**:
  - `method`: RPC method name (e.g., "create_kernel", "destroy_kernel", "execute_code")
- Tracks processing frequency by Manager command type

**`backendai_rpc_failure_requests`** (Counter)
- **Description**: Failed RPC request count
- **Labels**:
  - `method`: RPC method name that failed
  - `exception`: Exception class name (e.g., "ContainerNotFound", "ResourceExhausted")
- Analyzes RPC command processing failure causes

**`backendai_rpc_request_duration_seconds`** (Histogram)
- **Description**: RPC request processing time in seconds
- **Labels**:
  - `method`: RPC method name
- **Buckets**: [0.001, 0.01, 0.1, 0.5, 1, 2, 5, 10, 30, 60] seconds
- Measures performance of RPC commands like kernel creation/deletion

#### Resource Utilization Metrics
Metrics related to container and hardware resource usage.

**`backendai_container_utilization`** (Gauge)
- **Description**: Container resource utilization per kernel
- **Labels**:
  - `container_metric_name`: Metric type (cpu_used, cpu_util, mem, cuda_mem, cuda_util, net_rx, net_tx)
  - `agent_id`: Agent identifier
  - `kernel_id`: Kernel identifier (Backend.AI's container wrapper, not the actual container ID)
  - `session_id`: Session identifier
  - `user_id`: User UUID who owns the session
  - `project_id`: Project UUID that the session belongs to
  - `value_type`: Value interpretation (current, capacity, pct)
- Tracks resource usage per kernel (CPU, memory, GPU, network, etc.)

**`backendai_device_utilization`** (Gauge)
- **Description**: Physical device utilization on the agent node
- **Labels**:
  - `device_metric_name`: Metric type (cpu_util, mem, cuda_mem, cuda_util, cuda_temp, cuda_power)
  - `agent_id`: Agent identifier
  - `device_id`: Hardware device identifier (e.g., GPU index)
  - `value_type`: Value interpretation (current, capacity, pct)
- Tracks hardware device usage (GPU, CPU, memory) on the agent node

#### Container Lifecycle Synchronization Metrics
Metrics related to container state synchronization operations between Manager and Agent.

**`backendai_sync_container_lifecycle_trigger_count`** (Counter)
- **Description**: Number of times container lifecycle sync was triggered
- **Labels**:
  - `agent_id`: Agent identifier
- Frequency of container state synchronization between Manager and Agent

**`backendai_sync_container_lifecycle_success_count`** (Counter)
- **Description**: Number of successfully synchronized containers
- **Labels**:
  - `agent_id`: Agent identifier
- Number of successfully synchronized kernels

**`backendai_sync_container_lifecycle_failure_count`** (Counter)
- **Description**: Number of failed container synchronization attempts
- **Labels**:
  - `agent_id`: Agent identifier
  - `exception`: Exception class name causing the failure
- Tracks synchronization failure causes

#### Statistics Collection Task Metrics
Metrics related to node, container, and process level statistics collection.

**`backendai_stat_task_trigger_count`** (Counter)
- **Description**: Number of times statistics collection was triggered
- **Labels**:
  - `agent_id`: Agent identifier
  - `stat_scope`: Statistics scope (node, container, process)
- Collection frequency by Node/Container/Process level

**`backendai_stat_task_success_count`** (Counter)
- **Description**: Number of successful statistics collection operations
- **Labels**:
  - `agent_id`: Agent identifier
  - `stat_scope`: Statistics scope (node, container, process)
- Tracks statistics collection success rate

**`backendai_stat_task_failure_count`** (Counter)
- **Description**: Number of failed statistics collection operations
- **Labels**:
  - `agent_id`: Agent identifier
  - `stat_scope`: Statistics scope (node, container, process)
  - `exception`: Exception class name causing the failure
- Analyzes statistics collection failure causes

### Prometheus Query Examples

The following examples demonstrate common Prometheus queries for Agent metrics. Note that Counter metrics use the `_total` suffix and Histogram metrics use `_bucket`, `_sum`, `_count` suffixes in actual queries.

**Important Notes:**
- When using `increase()` or `rate()` functions, the time range must be at least 2-4x longer than your Prometheus scrape interval to get reliable data. If the time range is too short, metrics may not appear or show incomplete data.
- Default Prometheus scrape interval is typically 15s-30s
- **Time range selection trade-offs**:
  - Shorter ranges (e.g., `[1m]`): Detect changes faster with more granular data, but more sensitive to noise and short-term fluctuations
  - Longer ranges (e.g., `[5m]`): Smoother graphs with reduced noise, better for identifying trends, but slower to detect sudden changes
  - For real-time alerting: Use shorter ranges like `[1m]` or `[2m]`
  - For dashboards and trend analysis: Use longer ranges like `[5m]` or `[10m]`

#### RPC Request Monitoring

**RPC Request Rate by Method**

Monitor RPC request rate by method. This shows how frequently the Manager sends commands to agents. Use this to understand agent workload and command distribution.

```promql
sum(rate(backendai_rpc_requests_total{service_group="$service_groups"}[5m])) by (method)
```

**RPC Failure Rate by Method**

Track RPC failure rate by method and exception type. This helps identify which RPC operations are failing and why.

```promql
sum(rate(backendai_rpc_failure_requests_total{service_group="$service_groups"}[5m])) by (method, exception)
```

**P95 RPC Request Duration**

Calculate P95 RPC request duration by method. This shows how long critical operations like kernel creation take. Use this to identify performance bottlenecks in agent operations.

```promql
histogram_quantile(0.95,
  sum(rate(backendai_rpc_request_duration_seconds_bucket{service_group="$service_groups"}[5m])) by (le, method)
)
```

#### Container Resource Utilization

**Top 5 GPU Memory Usage by Kernel**

Monitor GPU memory usage by kernel (top 5). This identifies kernels consuming the most GPU memory. Note: `topk` returns the top N items at each evaluation time. Over multiple time points, you may see more than 5 unique kernels if different kernels appear in the top 5 at different times.

```promql
topk(5, sum(backendai_container_utilization{container_metric_name="cuda_mem", value_type="current"}) by (kernel_id))
```

**Top 5 GPU Utilization by Kernel**

Monitor GPU utilization by kernel (top 5). This shows which kernels are actively using GPU compute. Note: `topk` returns the top N items at each evaluation time. Over multiple time points, you may see more than 5 unique kernels if different kernels appear in the top 5 at different times.

```promql
topk(5, sum(backendai_container_utilization{container_metric_name="cuda_util", value_type="current"}) by (kernel_id))
```

**Top 5 CPU Utilization Increase by Kernel**

Track CPU utilization increase rate by kernel (top 5). This shows CPU consumption trends over time. Note: Divided by 1000 to convert from per-mille to percentage. `topk` returns the top N items at each evaluation time. Over multiple time points, you may see more than 5 unique kernels if different kernels appear in the top 5 at different times.

```promql
topk(5, sum(increase(backendai_container_utilization{container_metric_name="cpu_util", value_type="current"}[5m])) by (kernel_id)) / 1000
```

**Network RX Traffic by Kernel**

Monitor network RX (receive) traffic by kernel. This tracks inbound network bandwidth usage per container.

```promql
sum(increase(backendai_container_utilization{container_metric_name="net_rx", value_type="current"}[5m])) by (kernel_id)
```

**Network TX Traffic by Kernel**

Monitor network TX (transmit) traffic by kernel. This tracks outbound network bandwidth usage per container.

```promql
sum(increase(backendai_container_utilization{container_metric_name="net_tx", value_type="current"}[5m])) by (kernel_id)
```

#### Device Utilization (Agent-level)

**GPU Memory Usage by Agent**

Monitor GPU memory usage at the agent level. This shows total GPU memory consumption on each agent node.

```promql
sum(backendai_device_utilization{device_metric_name="cuda_mem", value_type="current"}) by (agent_id)
```

**GPU Memory Capacity by Agent**

Monitor GPU memory capacity. This shows the maximum available GPU memory on each agent.

```promql
sum(backendai_device_utilization{device_metric_name="cuda_mem", value_type="capacity"}) by (agent_id)
```

**GPU Utilization Percentage by Agent**

Monitor GPU utilization percentage at the agent level. This shows overall GPU usage across all kernels on the agent.

```promql
sum(backendai_device_utilization{device_metric_name="cuda_util", value_type="current"}) by (agent_id)
```

**CPU Utilization Increase by Agent**

Track CPU utilization increase rate at the agent level. This shows agent-wide CPU consumption trends.

```promql
sum(increase(backendai_device_utilization{device_metric_name="cpu_util", value_type="current"}[5m])) by (agent_id) / 1000
```

**Network RX Traffic by Agent**

Monitor network RX (receive) traffic at the agent level. This tracks total inbound network bandwidth usage on each agent node.

```promql
sum(increase(backendai_device_utilization{device_metric_name="net_rx", value_type="current"}[5m])) by (agent_id)
```

**Network TX Traffic by Agent**

Monitor network TX (transmit) traffic at the agent level. This tracks total outbound network bandwidth usage on each agent node.

```promql
sum(increase(backendai_device_utilization{device_metric_name="net_tx", value_type="current"}[5m])) by (agent_id)
```

#### Container Lifecycle Synchronization

**Container Sync Trigger Frequency**

Monitor container sync trigger frequency. This shows how often the agent synchronizes container state with the Manager.

```promql
sum(rate(backendai_sync_container_lifecycle_trigger_count_total[5m])) by (agent_id)
```

**Container Sync Success Rate**

Track container sync success rate. This helps ensure containers are properly synchronized.

```promql
sum(rate(backendai_sync_container_lifecycle_success_count_total[5m])) by (agent_id)
```

**Container Sync Failure Rate**

Monitor container sync failures. This identifies issues with container state synchronization.

```promql
sum(rate(backendai_sync_container_lifecycle_failure_count_total[5m])) by (agent_id, exception)
```

#### Statistics Collection

**Statistics Collection Trigger Rate**

Monitor statistics collection trigger rate by scope. This shows how frequently statistics are collected at different levels.

```promql
sum(rate(backendai_stat_task_trigger_count_total[5m])) by (agent_id, stat_scope)
```

**Statistics Collection Success Rate**

Track statistics collection success rate. This ensures statistics collection is working properly.

```promql
sum(rate(backendai_stat_task_success_count_total[5m])) by (agent_id, stat_scope)
```

**Statistics Collection Failure Rate**

Monitor statistics collection failures. This identifies issues with resource monitoring.

```promql
sum(rate(backendai_stat_task_failure_count_total[5m])) by (agent_id, stat_scope, exception)
```

### Logs
Agent logs include:
- Kernel creation and termination events
- Resource allocation and release
- Error situations and recovery operations
- RPC request/response tracking

## Development

See [README.md](./README.md) for development setup instructions.

## Related Documentation

- [Manager Component](../manager/README.md) - Session orchestration
- [Storage Proxy Component](../storage/README.md) - Virtual folder management
- [Overall Architecture](../README.md) - System-wide architecture
""",
    'long_description_content_type': 'text/markdown',
    'name': 'backend.ai-agent',
    'namespace_packages': (
    ),
    'package_data': {
        'ai.backend.agent': (
            'VERSION',
            'py.typed',
        ),
        'ai.backend.agent.docker': (
            'backendai-socket-relay.img.aarch64.tar.gz',
            'backendai-socket-relay.img.x86_64.tar.gz',
            'backendai-socket-relay.version.txt',
            'linuxkit-metadata-proxy-worker.aarch64.bin',
            'linuxkit-metadata-proxy-worker.x86_64.bin',
            'linuxkit-nsenter.img.aarch64.tar.gz',
            'linuxkit-nsenter.img.x86_64.tar.gz',
            'linuxkit-nsenter.version.txt',
        ),
    },
    'packages': (
        'ai.backend.agent',
        'ai.backend.agent.cli',
        'ai.backend.agent.config',
        'ai.backend.agent.docker',
        'ai.backend.agent.docker.metadata',
        'ai.backend.agent.dummy',
        'ai.backend.agent.kernel_registry',
        'ai.backend.agent.kubernetes',
        'ai.backend.agent.metrics',
        'ai.backend.agent.observer',
        'ai.backend.agent.plugin',
        'ai.backend.agent.stage',
        'ai.backend.agent.stage.kernel_lifecycle',
        'ai.backend.agent.stage.kernel_lifecycle.docker',
        'ai.backend.agent.stage.kernel_lifecycle.docker.mount',
        'ai.backend.agent.vendor',
        'ai.backend.agent.watcher',
    ),
    'project_urls': {
        'Documentation': 'https://docs.backend.ai/',
        'Source': 'https://github.com/lablup/backend.ai',
    },
    'python_requires': '>=3.13,<3.14',
    'url': 'https://github.com/lablup/backend.ai',
    'version': """25.16.0
""",
    'zip_safe': False,
})



import asyncio
from decimal import Decimal
from pathlib import Path
from typing import (
    Any,
    FrozenSet,
    Literal,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Tuple,
    override,
)

from ai.backend.common.config import read_from_file
from ai.backend.common.docker import ImageRef
from ai.backend.common.dto.agent.response import PurgeImagesResp
from ai.backend.common.dto.manager.rpc_request import PurgeImagesReq
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.types import (
    AutoPullBehavior,
    ClusterInfo,
    ClusterSSHPortMapping,
    ContainerId,
    ContainerStatus,
    DeviceId,
    DeviceName,
    ImageConfig,
    ImageRegistry,
    KernelCreationConfig,
    KernelId,
    MountPermission,
    MountTypes,
    ResourceSlot,
    Sentinel,
    ServicePort,
    SessionId,
    SlotName,
    current_resource_slots,
)

from ..agent import (
    ACTIVE_STATUS_SET,
    AbstractAgent,
    AbstractKernelCreationContext,
    ScanImagesResult,
)
from ..config.unified import AgentUnifiedConfig
from ..exception import UnsupportedResource
from ..kernel import AbstractKernel
from ..resources import (
    AbstractComputePlugin,
    ComputerContext,
    KernelResourceSpec,
    Mount,
    known_slot_types,
)
from ..types import Container, KernelOwnershipData, MountInfo
from .config import DEFAULT_CONFIG_PATH, dummy_local_config
from .kernel import DummyKernel
from .resources import load_resources, scan_available_resources


class DummyKernelCreationContext(AbstractKernelCreationContext[DummyKernel]):
    dummy_config: Mapping[str, Any]

    def __init__(
        self,
        ownership_data: KernelOwnershipData,
        event_producer: EventProducer,
        kenrel_image: ImageRef,
        kernel_config: KernelCreationConfig,
        distro: str,
        local_config: AgentUnifiedConfig,
        computers: MutableMapping[DeviceName, ComputerContext],
        restarting: bool = False,
        *,
        dummy_config: Mapping[str, Any],
    ) -> None:
        super().__init__(
            ownership_data,
            event_producer,
            kenrel_image,
            kernel_config,
            distro,
            local_config,
            computers,
            restarting=restarting,
        )
        self.dummy_config = dummy_config
        self.creation_ctx_config = self.dummy_config["kernel-creation-ctx"]

    async def get_extra_envs(self) -> Mapping[str, str]:
        return {}

    async def prepare_resource_spec(
        self,
    ) -> Tuple[KernelResourceSpec, Optional[Mapping[str, Any]]]:
        slots = ResourceSlot.from_json(self.kernel_config["resource_slots"])
        # Ensure that we have intrinsic slots.
        assert SlotName("cpu") in slots
        assert SlotName("mem") in slots
        # accept unknown slot type with zero values
        # but reject if they have non-zero values.
        for st, sv in slots.items():
            if st not in known_slot_types and sv != Decimal(0):
                raise UnsupportedResource(st)
        # sanitize the slots
        current_resource_slots.set(known_slot_types)
        slots = slots.normalize_slots(ignore_unknown=True)
        resource_spec = KernelResourceSpec(
            allocations={},
            slots={**slots},  # copy
            mounts=[],
            scratch_disk_size=0,  # TODO: implement (#70)
        )
        resource_opts = self.kernel_config.get("resource_opts", {})
        return resource_spec, resource_opts

    async def prepare_scratch(self) -> None:
        delay = self.creation_ctx_config["delay"]["prepare-scratch"]
        await asyncio.sleep(delay)

    async def get_intrinsic_mounts(self) -> Sequence[Mount]:
        return []

    @property
    @override
    def repl_ports(self) -> Sequence[int]:
        return (2000, 2001)

    @property
    @override
    def protected_services(self) -> Sequence[str]:
        return ()

    async def apply_network(self, cluster_info: ClusterInfo) -> None:
        return

    async def prepare_ssh(self, cluster_info: ClusterInfo) -> None:
        delay = self.creation_ctx_config["delay"]["prepare-ssh"]
        await asyncio.sleep(delay)

    async def process_mounts(self, mounts: Sequence[Mount]):
        return

    async def apply_accelerator_allocation(
        self,
        computer: AbstractComputePlugin,
        device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> None:
        return

    async def generate_accelerator_mounts(
        self,
        computer: AbstractComputePlugin,
        device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> list[MountInfo]:
        return []

    def resolve_krunner_filepath(self, filename) -> Path:
        return Path()

    def get_runner_mount(
        self,
        type: MountTypes,
        src: str | Path,
        target: str | Path,
        perm: MountPermission = MountPermission.READ_ONLY,
        opts: Optional[Mapping[str, Any]] = None,
    ):
        return Mount(MountTypes.BIND, Path(), Path())

    async def prepare_container(
        self,
        resource_spec: KernelResourceSpec,
        environ: Mapping[str, str],
        service_ports,
        cluster_info: ClusterInfo,
    ) -> DummyKernel:
        delay = self.creation_ctx_config["delay"]["spawn"]
        await asyncio.sleep(delay)
        return DummyKernel(
            self.ownership_data,
            self.kernel_config["network_id"],
            self.image_ref,
            self.kspec_version,
            agent_config=self.local_config.model_dump(by_alias=True),
            service_ports=service_ports,
            resource_spec=resource_spec,
            environ=environ,
            data={},
            dummy_config=self.dummy_config,
        )

    async def start_container(
        self,
        kernel_obj: AbstractKernel,
        cmdargs: list[str],
        resource_opts,
        preopen_ports,
        cluster_info: ClusterInfo,
    ) -> Mapping[str, Any]:
        container_bind_host = self.local_config.container.bind_host
        advertised_kernel_host = self.local_config.container.advertised_host
        delay = self.creation_ctx_config["delay"]["start-container"]
        await asyncio.sleep(delay)
        return {
            "container_id": "",
            "kernel_host": advertised_kernel_host or container_bind_host,
            "repl_in_port": 2000,
            "repl_out_port": 2001,
            "stdin_port": 2002,  # legacy
            "stdout_port": 2003,  # legacy
            "host_ports": [2000, 2001, 2002, 2003],
            "domain_socket_proxies": [],
            "block_service_ports": self.internal_data.get("block_service_ports", False),
        }

    async def mount_krunner(
        self,
        resource_spec: KernelResourceSpec,
        environ: MutableMapping[str, str],
    ) -> None:
        delay = self.creation_ctx_config["delay"]["mount-krunner"]
        await asyncio.sleep(delay)


class DummyAgent(
    AbstractAgent[DummyKernel, DummyKernelCreationContext],
):
    dummy_config: Mapping[str, Any]

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        raw_config, _ = read_from_file(DEFAULT_CONFIG_PATH, "dummy")
        self.dummy_config = dummy_local_config.check(raw_config)
        self.dummy_agent_cfg = self.dummy_config["agent"]

    async def execute(
        self,
        session_id: SessionId,
        kernel_id: KernelId,
        run_id: str | None,
        mode: Literal["query", "batch", "input", "continue"],
        text: str,
        *,
        opts: Mapping[str, Any],
        api_version: int,
        flush_timeout: float,
    ):
        return {"status": "not-implemented"}

    def get_public_service_ports(self, service_ports: list[ServicePort]) -> list[ServicePort]:
        return []

    async def sync_container_lifecycles(self, interval: float) -> None:
        return

    async def extract_command(self, image: str) -> str | None:
        return None

    async def enumerate_containers(
        self,
        status_filter: FrozenSet[ContainerStatus] = ACTIVE_STATUS_SET,
    ) -> Sequence[Tuple[KernelId, Container]]:
        return []

    async def resolve_image_distro(self, image: ImageConfig) -> str:
        return "ubuntu16.04"

    @override
    def get_cgroup_path(self, controller: str, container_id: str) -> Path:
        #  Dummy agent does not use cgroups, so we return an empty path.
        return Path()

    @override
    def get_cgroup_version(self) -> str:
        # Dummy agent does not use cgroups, so we return an empty string.
        return ""

    async def load_resources(self) -> Mapping[DeviceName, AbstractComputePlugin]:
        return await load_resources(
            self.etcd, self.local_config.model_dump(by_alias=True), self.dummy_config
        )

    async def scan_available_resources(self) -> Mapping[SlotName, Decimal]:
        return await scan_available_resources(
            self.local_config.model_dump(by_alias=True),
            {name: cctx.instance for name, cctx in self.computers.items()},
        )

    async def extract_image_command(self, image: str) -> str | None:
        delay = self.dummy_agent_cfg["delay"]["scan-image"]
        await asyncio.sleep(delay)
        return "cr.backend.ai/stable/python:3.9-ubuntu20.04"

    async def scan_images(self) -> ScanImagesResult:
        delay = self.dummy_agent_cfg["delay"]["scan-image"]
        await asyncio.sleep(delay)
        return ScanImagesResult(scanned_images={}, removed_images={})

    async def pull_image(
        self,
        image_ref: ImageRef,
        registry_conf: ImageRegistry,
        *,
        timeout: float | None,
    ) -> None:
        delay = self.dummy_agent_cfg["delay"]["pull-image"]
        await asyncio.sleep(delay)

    async def push_image(
        self,
        image_ref: ImageRef,
        registry_conf: ImageRegistry,
        *,
        timeout: float | None | Sentinel = Sentinel.TOKEN,
    ) -> None:
        delay = self.dummy_agent_cfg["delay"]["push-image"]
        await asyncio.sleep(delay)

    async def purge_images(self, request: PurgeImagesReq) -> PurgeImagesResp:
        delay = self.dummy_agent_cfg["delay"]["purge-images"]
        await asyncio.sleep(delay)
        return PurgeImagesResp([])

    async def check_image(
        self, image_ref: ImageRef, image_id: str, auto_pull: AutoPullBehavior
    ) -> bool:
        if (
            existing_imgs := self.dummy_agent_cfg["image"]["already-have"]
        ) is not None and image_ref in existing_imgs:
            return True
        return False

    async def init_kernel_context(
        self,
        ownership_data: KernelOwnershipData,
        kernel_image: ImageRef,
        kernel_config: KernelCreationConfig,
        *,
        restarting: bool = False,
        cluster_ssh_port_mapping: Optional[ClusterSSHPortMapping] = None,
    ) -> DummyKernelCreationContext:
        distro = await self.resolve_image_distro(kernel_config["image"])
        return DummyKernelCreationContext(
            ownership_data,
            self.event_producer,
            kernel_image,
            kernel_config,
            distro,
            self.local_config,
            self.computers,
            restarting=restarting,
            dummy_config=self.dummy_config,
        )

    async def destroy_kernel(
        self,
        kernel_id: KernelId,
        container_id: Optional[ContainerId],
    ) -> None:
        delay = self.dummy_agent_cfg["delay"]["destroy-kernel"]
        await asyncio.sleep(delay)

    async def clean_kernel(
        self,
        kernel_id: KernelId,
        container_id: Optional[ContainerId],
        restarting: bool,
    ) -> None:
        delay = self.dummy_agent_cfg["delay"]["clean-kernel"]
        await asyncio.sleep(delay)

    async def create_local_network(self, network_name: str) -> None:
        delay = self.dummy_agent_cfg["delay"]["create-network"]
        await asyncio.sleep(delay)

    async def destroy_local_network(self, network_name: str) -> None:
        delay = self.dummy_agent_cfg["delay"]["destroy-network"]
        await asyncio.sleep(delay)

    async def restart_kernel__load_config(
        self,
        kernel_id: KernelId,
        name: str,
    ) -> bytes:
        await asyncio.sleep(0.1)
        return b""

    async def restart_kernel__store_config(
        self,
        kernel_id: KernelId,
        name: str,
        data: bytes,
    ) -> None:
        await asyncio.sleep(0.1)
